# -*- coding: utf8 -*-
'''
Library for combined x-ray and neutrons simulations with adaptive layer segmentation
====================================================================================
Library for specular neutron and x-ray reflectometry of more complex structures where elemental composition
and/or magnetism is better described separately than within one slap model. The model sums up a set of
*Elements* to calculate a total SLD profile and then uses adaptive layer segmentation to model it.

The actual modeling of the result structure is done with the same function as in spec_nx.

Classes
-------

Layer
~~~~~
``Layer(b = 0.0, d = 0.0, f = 0.0+0.0J, b=0j, dens = 1.0, magn_ang = 0.0, magn = 0.0, sigma = 0.0, xs_ai=0.0, rough_type=0, sigma_magn=0.0, magn_void=False)``

``b``
   The neutron scattering length per formula unit in fm (fermi meter =
   1e-15m)
``d``
   The thickness of the layer in AA (Angstroms = 1e-10m)
``f``
   The x-ray scattering length per formula unit in electrons. To be
   strict it is the number of Thompson scattering lengths for each
   formula unit.
``dens``
   The density of formula units in units per Angstroms. Note the units!
``magn_ang``
   The angle of the magnetic moment in degress. 0 degrees correspond to
   a moment collinear with the neutron spin.
``magn``
   The magnetic moment per formula unit (same formula unit as b and dens
   refer to)
``sigma``
   The root mean square roughness of the top interface of the layer in
   Angstroms.
``sigma_mag``
   A different roughness parameter for the magnetization of the layer, 0 is ignored
``xs_ai``
   The sum of the absorption cross section and the incoherent scattering
   cross section in barns for neutrons
``magn_void``
   If true this layer has no magnetization. In case of *sigma_mag* beging larger then 0 the additional
   roughness is only applied to the magnetic layer and inside this layer follows the chemical profile.
``rough_type``
   Used model to get the SLD profile of the interface, *0* is an error function profile (gaussian roughness),
   *1* is a linear profile, *2* and *3* are exponential decays from bottom or top side.


Stack
~~~~~
``Stack(Layers = [], Repetitions = 1, Element = 0)``

``Layers``
   A ``list`` consiting of ``Layer``\ s in the stack the first item is
   the layer closest to the bottom
``Repetitions``
   The number of repsetions of the stack
``Element``
   The Element of the model that this stack belongs to. There has to be at least one stack with Element 0.
   For every *Element* the layers are stacked *on top* of the substrate separately and then all *Elements* are
   summed up to calculate the total SLD.
   The main use case for this is either to separate magnetic from nuclear structure
   (nuclear Element=0, magnetic Element=1) or two or more elemental contributions for element specific diffusion.
   For layers that have no contribution at a certain depth one can add a layer with 0 density as spacer.

Sample
~~~~~~
``Sample(Stacks = [], Ambient = Layer(), Substrate = Layer(), minimal_steps=0.5, max_diff_n=0.01, max_diff_x=0.01, smoothen=0)``

``Stacks``
   A ``list`` consiting of ``Stack``\ s in the stacks the first item is
   the layer closest to the bottom
``Ambient``
   A ``Layer`` describing the Ambient (enviroment above the sample).
   Only the scattering lengths and density of the layer is used.
``Substrate``
   A ``Layer`` describing the substrate (enviroment below the sample).
   Only the scattering lengths, density and roughness of the layer is
   used.
``minimal_steps``
   The thickness of the minimal step between layers. Smaller values make the model more precise but slower.
   For data with larger q-range a smaller minimal_step is required. Try to start with 0.5-1 Å step size and
   increase the value until you see differences in the simulated data.
``max_diff_n``
   Maximum neutron SLD deviation to be allowed for layers to be combined in the adaptive procedure
``max_diff_x``
   Maximum x-ray SLD deviation to be allowed for layers to be combined in the adaptive procedure
``smoothen``
   Default is to not use any roughness for the segmentation. If set to 1 this will add roughnesses between all
   segments to make the curve more smooth.

Instrument
~~~~~~~~~~
``Instrument(probe = 'x-ray', wavelength = 1.54, coords = '2θ', I0 = 1.0 res = 0.001, restype = 'no conv', respoints = 5, resintrange = 2, beamw = 0.01, footype = 'no corr', samplelen = 10.0, incangle = 0.0, pol = 'uu')``

``probe``
    Describes the radiation and measurments used is one of: 'x-ray',
    'neutron', 'neutron pol', 'neutron pol spin flip', 'neutron tof',
    'neutron pol tof' or the respective number 0, 1, 2, 3, 4, 5, 6. The
    calculations for x-rays uses ``f`` for the scattering length for
    neutrons ``b`` for 'neutron pol', 'neutron pol spin flip' and 'neutron
    pol tof' alternatives the ``magn`` is used in the calculations. Note
    that the angle of magnetization ``magn_ang`` is only used in the last
    alternative.
``wavelength``
    The wavelength of the radiation given in AA (Angstroms)
``coords``
    The coordinates of the data given to the SimSpecular function. The
    available alternatives are: 'q' or '2θ'. Alternatively the numbers 0 (q)
    or 1 (tth) can be used.
``I0``
    The incident intensity (a scaling factor)
``Ibkg``
    The background intensity. Added as a constant value to the calculated
    reflectivity
``res``
    The resolution of the instrument given in the coordinates of ``coords``.
    This assumes a gaussian resolution function and ``res`` is the standard
    deviation of that gaussian. If ``restype`` has (dx/x) in its name the
    gaussian standard deviation is given by res*x where x is either in tth
    or q.
``restype``
    Describes the rype of the resolution calculated. One of the
    alterantives: 'no conv', 'fast conv', 'full conv and varying res.',
    'fast conv + varying res.', 'full conv and varying res. (dx/x)', 'fast
    conv + varying res. (dx/x)'. The respective numbers 0-3 also works. Note
    that fast convolution only alllows a single value into res wheras the
    other can also take an array with the same length as the x-data (varying
    resolution)
``respoints``
    The number of points to include in the resolution calculation. This is
    only used for 'full conv and vaying res.', 'fast conv + varying res',
    'full conv and varying res. (dx/x)' and 'fast conv + varying res.
    (dx/x)'.
``resintrange``
    Number of standard deviatons to integrate the resolution function times
    the reflectivity over
``footype``
    Which type of footprint correction is to be applied to the simulation.
    One of: 'no corr', 'gauss beam' or 'square beam'. Alternatively, the
    number 0-2 are also valid. The different choices are self expnalatory.
``beamw``
    The width of the beam given in mm. For 'gauss beam' it should be the
    standard deviation. For 'square beam' it is the full width of the beam.
``samplelen``
    The length of the sample given in mm
``incangle``
    The incident angle of the neutrons, only valid in tof mode
``pol``
    The measured polarization of the instrument. Valid options are:
    'uu','dd', 'ud', 'du' or 'ass' the respective number 0-3 also works.
'''
from copy import deepcopy
from numpy import *
from scipy.special import wofz
from . import spec_nx
from .spec_nx import refl
from .lib.instrument import *

# Preamble to define the parameters needed for the models outlined below:
ModelID='SpecAdaptive'
__pars__=spec_nx.__pars__.copy()

instrument_string_choices=spec_nx.instrument_string_choices.copy()
InstrumentParameters=spec_nx.InstrumentParameters.copy()
InstrumentGroups=spec_nx.InstrumentGroups.copy()
InstrumentUnits=spec_nx.InstrumentUnits.copy()

LayerParameters=spec_nx.LayerParameters.copy()
LayerParameters.update({'magn_void': False, 'rough_type': 0, 'sigma_mag': 0.0})
LayerUnits=spec_nx.LayerUnits.copy()
LayerUnits.update({'magn_void': 'True/False',
                   'rough_type': '0-gauss/1-lin\n2/3-exp', 'sigma_mag': 'AA'})
LayerGroups=[('Standard', ['f', 'dens', 'd', 'sigma']),
             ('Neutron', ['b', 'xs_ai', 'magn', 'magn_ang', 'sigma_mag']),
             ('Special', ['magn_void', 'rough_type'])]

StackParameters={'Layers': [], 'Repetitions': 1, 'Element': 0}
SampleParameters=spec_nx.SampleParameters.copy()
SampleParameters.update({'minimal_steps': 0.5, 'max_diff_n': 0.01,
                         'max_diff_x': 0.01, 'smoothen': 0})
sample_string_choices={}

AA_to_eV=spec_nx.AA_to_eV
q_limit=spec_nx.q_limit
Buffer=spec_nx.Buffer

def SLD_calculations(z, item, sample, inst):
    res=spec_nx.SLD_calculations(z, item, sample, inst)
    res['z']-=5*refl.resolve_par(sample.Substrate, 'sigma')
    return res

SimulationFunctions={'Specular': spec_nx.Specular,
                     'SLD': SLD_calculations
                     }

(Instrument, Layer, Stack, Sample)=refl.MakeClasses(InstrumentParameters,
                                                    LayerParameters, StackParameters, SampleParameters,
                                                    SimulationFunctions,
                                                    ModelID)

def calculate_segmentation(sample):
    '''
      Calculate segmentation steps inside a sample defined by
      a maximum SLD slope and minimum step size. It first
      calculates the nuclear and magnetic SLD profile
      from the model and than separates it according to the
      given parameters.
    '''
    parameters=resolve_parameters_by_element(sample)
    dens=array(parameters['dens'], dtype=float64)
    f=array(parameters['f'], dtype=complex128)
    b=array(parameters['b'], dtype=complex128)
    xs_ai=array(parameters['xs_ai'], dtype=float64)
    magn=array(parameters['magn'], dtype=float64)
    magn_ang=array(parameters['magn_ang'], dtype=float64)/180.*pi
    magn_void=array(parameters['magn_void'], dtype=float64)

    sld_x=dens*f
    sld_n=dens*b
    sld_xs=dens*xs_ai
    # the new magnetization density and angle will be
    # calculated from perpendicular and parallel components to allow
    # a smooth transition in magnetic angle
    mag_sld_nsf=magn*dens*cos(magn_ang)
    mag_sld_sf=magn*dens*sin(magn_ang)

    d=array(parameters['d'], dtype=float64)
    d=d[1:-1]
    # Include one extra element - the zero pos (substrate/film interface)
    int_pos=cumsum(r_[0, d])
    rough_type=array(parameters['rough_type'], dtype=int)[:-1]
    sigma_n=array(parameters['sigma'], dtype=float64)[:-1]+1e-7
    sigma_m=array(parameters['sigma_mag'], dtype=float64)[:-1]+1e-7
    z=arange(-sigma_n[0]*5, int_pos.max()+sigma_n[-1]*5, sample.minimal_steps/5.0)
    # interface transition functions
    trans_n=(0.5-0.5*erf((z[:, newaxis]-int_pos)/sqrt(2.)/sigma_n))*(rough_type==0)
    trans_n+=maximum(0., minimum(1., (1.+(int_pos-z[:, newaxis])/2./sigma_n)/2.))*(rough_type==1)
    trans_n+=maximum(0., minimum(1., exp((int_pos-z[:, newaxis])/sigma_n)))*(rough_type==2)
    trans_n+=maximum(0., minimum(1., 1.-exp((z[:, newaxis]-int_pos)/sigma_n)))*(rough_type==3)
    trans_m=(0.5-0.5*erf((z[:, newaxis]-int_pos)/sqrt(2.)/sigma_m))*(rough_type==0)
    trans_m+=maximum(0., minimum(1., (1.+(int_pos-z[:, newaxis])/2./sigma_m)/2.))*(rough_type==1)
    trans_m+=maximum(0., minimum(1., exp((int_pos-z[:, newaxis])/sigma_m)))*(rough_type==2)
    trans_m+=maximum(0., minimum(1., 1.-exp((z[:, newaxis]-int_pos)/sigma_m)))*(rough_type==3)
    # SLD calculations
    rho_x=sum((sld_x[:-1]-sld_x[1:])*trans_n, 1)+sld_x[-1]
    rho_n=sum((sld_n[:-1]-sld_n[1:])*trans_n, 1)+sld_n[-1]
    rho_m_nsf=sum((mag_sld_nsf[:-1]-mag_sld_nsf[1:])*trans_m, 1)+mag_sld_nsf[-1]
    rho_m_sf=sum((mag_sld_sf[:-1]-mag_sld_sf[1:])*trans_m, 1)+mag_sld_sf[-1]
    rho_void=sum((magn_void[:-1]-magn_void[1:])*trans_n, 1)+magn_void[-1]
    xs_ai_comb=sum((sld_xs[:-1]-sld_xs[1:])*trans_n, 1)+sld_xs[-1]
    # add more elements to the SLDs
    for params in parameters['Elements'][1:]:
        dens=array(params['dens'], dtype=float64)
        f=array(params['f'], dtype=complex128)
        b=array(params['b'], dtype=complex128)
        xs_ai=array(params['xs_ai'], dtype=float64)
        magn=array(params['magn'], dtype=float64)
        magn_ang=array(params['magn_ang'], dtype=float64)/180.*pi
        sld_x=dens*f
        sld_n=dens*b
        sld_xs=dens*xs_ai
        mag_sld_nsf=magn*dens*cos(magn_ang)
        mag_sld_sf=magn*dens*sin(magn_ang)

        d=array(params['d'], dtype=float64)
        d=d[1:-1]
        # Include one extra element - the zero pos (substrate/film interface)
        int_pos=cumsum(r_[0, d])
        rough_type=array(params['rough_type'], dtype=int)[:-1]
        sigma_n=array(params['sigma'], dtype=float64)[:-1]+1e-7
        sigma_m=array(params['sigma_mag'], dtype=float64)[:-1]+1e-7
        # interface transition functions
        trans_n=(0.5-0.5*erf((z[:, newaxis]-int_pos)/sqrt(2.)/sigma_n))*(rough_type==0)
        trans_n+=maximum(0., minimum(1., (1.+(int_pos-z[:, newaxis])/2./sigma_n)/2.))*(rough_type==1)
        trans_n+=maximum(0., minimum(1., exp((int_pos-z[:, newaxis])/sigma_n)))*(rough_type==2)
        trans_n+=maximum(0., minimum(1., 1.-exp((z[:, newaxis]-int_pos)/sigma_n)))*(rough_type==3)
        trans_m=(0.5-0.5*erf((z[:, newaxis]-int_pos)/sqrt(2.)/sigma_m))*(rough_type==0)
        trans_m+=maximum(0., minimum(1., (1.+(int_pos-z[:, newaxis])/2./sigma_m)/2.))*(rough_type==1)
        trans_m+=maximum(0., minimum(1., exp((int_pos-z[:, newaxis])/sigma_m)))*(rough_type==2)
        trans_m+=maximum(0., minimum(1., 1.-exp((z[:, newaxis]-int_pos)/sigma_m)))*(rough_type==3)
        # SLD calculations
        rho_x+=sum((sld_x[:-1]-sld_x[1:])*trans_n, 1)+sld_x[-1]
        rho_n+=sum((sld_n[:-1]-sld_n[1:])*trans_n, 1)+sld_n[-1]
        rho_m_nsf+=sum((mag_sld_nsf[:-1]-mag_sld_nsf[1:])*trans_m, 1)+mag_sld_nsf[-1]
        rho_m_sf+=sum((mag_sld_sf[:-1]-mag_sld_sf[1:])*trans_m, 1)+mag_sld_sf[-1]
        xs_ai_comb+=sum((sld_xs[:-1]-sld_xs[1:])*trans_n, 1)+sld_xs[-1]
    # calculate the segmentation
    d_segments=[0.]
    i=0
    rho_x_out=[rho_x[0]]
    rho_n_out=[rho_n[0]]
    rho_nsf_out=[rho_m_nsf[0]]
    rho_sf_out=[rho_m_sf[0]]
    xs_ai_out=[xs_ai_comb[0]]
    rho_x_r=rho_x.real
    rho_n_p=rho_n.real+rho_m_nsf
    rho_n_m=rho_n.real-rho_m_nsf
    while i<(len(z)-1):
        j=next_adaptive_segment(i, rho_x_r, rho_n_p, rho_n_m, rho_m_sf,
                                sample.max_diff_n, sample.max_diff_x, z)
        d_segments.append(z[j]-z[i])
        rho_x_out.append(rho_x[i:j].mean())
        rho_n_out.append(rho_n[i:j].mean())
        rho_nsf_out.append(rho_m_nsf[i:j].mean()*(1.-rho_void[i:j].mean()))  # averadge magn taking voids into account
        rho_sf_out.append(rho_m_sf[i:j].mean()*(1.-rho_void[i:j].mean()))  # averadge magn taking voids into account
        xs_ai_out.append(xs_ai_comb[i:j].mean())  # averadge mang angle
        i=j
    rho_nsf_out=array(rho_nsf_out)
    rho_sf_out=array(rho_sf_out)
    rho_m_out=sqrt(rho_nsf_out**2+rho_sf_out**2).tolist()
    magn_ang_out=(arctan2(rho_nsf_out, -rho_sf_out)*180./pi-90.).tolist()
    return (d_segments[1:], rho_x_out[1:],
            rho_n_out[1:], rho_m_out[1:], xs_ai_out[1:], magn_ang_out[1:])

def next_adaptive_segment(i, rho_x_r, rho_n_p, rho_n_m, rho_m_sf, max_diff_n, max_diff_x, z):
    # calculate the maximum variation of SLD up to given index
    diff_x=abs(maximum.accumulate(rho_x_r[i+1:])-minimum.accumulate(rho_x_r[i+1:]))
    diff_n_p=abs(maximum.accumulate(rho_n_p[i+1:])-minimum.accumulate(rho_n_p[i+1:]))
    diff_n_m=abs(maximum.accumulate(rho_n_m[i+1:])-minimum.accumulate(rho_n_m[i+1:]))
    diff_m_sf=abs(maximum.accumulate(rho_m_sf[i+1:])-minimum.accumulate(rho_m_sf[i+1:]))
    diff_idx=where(logical_not((diff_n_p<max_diff_n) & (diff_x<max_diff_x) &
                               (diff_n_m<max_diff_n) & (diff_m_sf<max_diff_n)))[0]
    if len(diff_idx)>0:
        j=min(len(z)-1, max(i+diff_idx[0]+1, i+5))
    else:
        j=len(z)-1  # last position
    return j

def resolve_parameters_by_element(sample):
    """
      Resolve the model standard parameters for each element.
      The first element used for normal parameter names,
      every other element does ignore substrate and ambience sld
      and is assigned to the 'elements' keyword as list.
      This makes it possible to build SLD profiles as sum of all
      elements.
    """
    elements=list(set([stack.Element for stack in sample.Stacks]))
    elements.sort()
    par=sample.Substrate._parameters.copy()
    for k in par:
        par[k]=[refl.resolve_par(sample.Substrate, k)]
    for k in sample.Substrate._parameters:
        for stack in sample.Stacks:
            if stack.Element!=0:
                continue
            par[k]=par[k]+stack.resolveLayerParameter(k)
        par[k]=par[k]+[refl.resolve_par(sample.Ambient, k)]
    par['sigma_mag']=where(array(par['sigma_mag'])!=0.,
                           par['sigma_mag'], par['sigma']).tolist()
    output=par
    output['Elements']=[]
    for element in elements:
        par=sample.Substrate._parameters.copy()
        for k in list(par.keys()):
            par[k]=[refl.resolve_par(sample.Substrate, k)]
        # zero substrat SLD
        par['f']=[0j]
        par['b']=[0j]
        par['magn']=[0.]
        par['xs_ai']=[0.]
        par['magn_ang']=[0.]
        for k in sample.Substrate._parameters:
            for stack in sample.Stacks:
                if stack.Element!=element:
                    continue
                par[k]=par[k]+stack.resolveLayerParameter(k)
            par[k]=par[k]+[refl.resolve_par(sample.Ambient, k)+0.0]
        # zero ambience SLD for Element
        par['f'][-1]=0j
        par['b'][-1]=0j
        par['magn'][-1]=0.
        par['xs_ai'][-1]=0.
        par['magn_ang'][-1]=0.
        # if magnetic roughness is set to zero use structural roughness
        par['sigma_mag']=where(array(par['sigma_mag'])!=0.,
                               par['sigma_mag'], par['sigma']).tolist()
        output['Elements'].append(par)
    return output

def resolveLayerParameters(self):
    # resolve parameters by creating layers automatically
    # adapting changes to the SLDs
    par={}
    for k in ['b', 'd', 'dens', 'f', 'magn', 'magn_ang', 'sigma', 'xs_ai']:
        par[k]=[refl.resolve_par(self.Substrate, k)]
    par['sigma'][0]=0.
    d, rho_x, rho_n, rho_m, xs_ai, magn_ang=calculate_segmentation(self)
    par['d']+=d
    par['f']+=rho_x
    par['b']+=rho_n
    par['magn']+=rho_m
    par['xs_ai']+=xs_ai
    if self.smoothen:
        par['sigma']+=[(d[i]+d[i+1])/4. for i in range(len(d)-1)]+[0.]
    else:
        par['sigma']+=[0. for ignore in d]
    par['magn_ang']+=magn_ang
    par['dens']+=[1. for ignore in d]
    for k in ['b', 'd', 'dens', 'f', 'magn', 'magn_ang', 'sigma', 'xs_ai']:
        par[k].append(refl.resolve_par(self.Ambient, k))
    return par

Sample.resolveLayerParameters=resolveLayerParameters

if __name__=='__main__':
    pass



import wx
import math
import os

from wx.lib.intctrl import IntCtrl

import string

from . import reflectivity_images as images
from genx.core.custom_logging import iprint

def is_reflfunction(obj):
    ''' Convenience function to determine whether obj belongs to the ReflFunction class.
    Return boolean.
    '''
    return obj.__class__.__name__=='ReflFunction'

class TextObjectValidator(wx.Validator):
    """ This validator is used to ensure that the user has entered something
        into the text object editor dialog's text field.
    """

    def __init__(self):
        """ Standard constructor.
        """
        wx.Validator.__init__(self)

    def Clone(self):
        """ Standard cloner.

           Note that every validator must implement the Clone() method.
        """
        return TextObjectValidator()

    def Validate(self, win):
        """ Validate the contents of the given text control.
        """
        textCtrl=self.GetWindow()
        text=textCtrl.GetValue()

        if len(text)==0:
            wx.MessageBox("A text object must contain some text!", "Error")
            textCtrl.SetBackgroundColour("pink")
            textCtrl.SetFocus()
            textCtrl.Refresh()
            return False
        else:
            textCtrl.SetBackgroundColour(
                wx.SystemSettings.GetColour(wx.SYS_COLOUR_WINDOW))
            textCtrl.Refresh()
            return True

    def TransferToWindow(self):
        """ Transfer data from validator to window.

            The default implementation returns False, indicating that an error
            occurred.  We simply return True, as we don't do any data transfer.
        """
        return True  # Prevent wxDialog from complaining.

    def TransferFromWindow(self):
        """ Transfer data from window to validator.
            The default implementation returns False, indicating that an error
            occurred.  We simply return True, as we don't do any data transfer.
         """
        return True  # Prevent wxDialog from complaining.

class MatchTextObjectValidator(wx.Validator):
    """ This validator is used to ensure that the user has entered something
        into the text object editor dialog's text field.
    """

    def __init__(self, stringlist):
        """ Standard constructor.
        """
        wx.Validator.__init__(self)
        self.stringlist=stringlist

    def Clone(self):
        """ Standard cloner.

            Note that every validator must implement the Clone() method.
        """
        return MatchTextObjectValidator(self.stringlist)

    def Validate(self, win):
        """ Validate the contents of the given text control.
        """
        textCtrl=self.GetWindow()
        text=textCtrl.GetValue()

        if self.stringlist.__contains__(text):
            textCtrl.SetBackgroundColour(
                wx.SystemSettings.GetColour(wx.SYS_COLOUR_WINDOW))
            textCtrl.Refresh()
            return True
        else:
            wx.MessageBox("The name is not defined!", "Error")
            textCtrl.SetBackgroundColour("pink")
            textCtrl.SetFocus()
            textCtrl.Refresh()
            return False

    def TransferToWindow(self):
        """ Transfer data from validator to window.
            The default implementation returns False, indicating that an error
            occurred.  We simply return True, as we don't do any data transfer.
        """
        return True  # Prevent wxDialog from complaining.

    def TransferFromWindow(self):
        """ Transfer data from window to validator.

            The default implementation returns False, indicating that an error
            occurred.  We simply return True, as we don't do any data transfer.
        """
        return True  # Prevent wxDialog from complaining.

class NoMatchTextObjectValidator(wx.Validator):
    """ This validator is used to ensure that the user has entered something
        into the text object editor dialog's text field.
    """

    def __init__(self, stringlist):
        """ Standard constructor.
        """
        wx.Validator.__init__(self)
        self.stringlist=stringlist

    def Clone(self):
        """ Standard cloner.
            Note that every validator must implement the Clone() method.
        """
        return NoMatchTextObjectValidator(self.stringlist)

    def Validate(self, win):
        """ Validate the contents of the given text control.
        """
        textCtrl=self.GetWindow()
        text=textCtrl.GetValue()
        if len(text)==0:
            wx.MessageBox("A text object must contain some text!", "Error")
            textCtrl.SetBackgroundColour("pink")
            textCtrl.SetFocus()
            textCtrl.Refresh()
            return False
        elif self.stringlist.__contains__(text):
            wx.MessageBox("Duplicates are not allowed!", "Error")
            textCtrl.SetBackgroundColour("pink")
            textCtrl.SetFocus()
            textCtrl.Refresh()
            return False
        else:
            textCtrl.SetBackgroundColour(
                wx.SystemSettings.GetColour(wx.SYS_COLOUR_WINDOW))
            textCtrl.Refresh()
            return True

    def TransferToWindow(self):
        """ Transfer data from validator to window.
            The default implementation returns False, indicating that an error
            occurred.  We simply return True, as we don't do any data transfer.
        """
        return True  # Prevent wxDialog from complaining.

    def TransferFromWindow(self):
        """ Transfer data from window to validator.
            The default implementation returns False, indicating that an error
            occurred.  We simply return True, as we don't do any data transfer.
        """
        return True  # Prevent wxDialog from complaining.

class NoMatchValidTextObjectValidator(wx.Validator):
    """ This validator is used to ensure that the user has entered something
        into the text object editor dialog's text field. It should not match
        a name in stringlist and it should be a valid varaible name. I.e 
        is should start with a letter and can only contains ordinary
        letters as in string.letters as well as numbers as in string.digits
        or _
    """

    def __init__(self, stringlist):
        """ Standard constructor.
        """
        wx.Validator.__init__(self)
        self.stringlist=stringlist
        self.reserved_words=['and', 'del', 'from', 'not', 'while', 'as',
                             'elif', 'global', 'or', 'with', 'assert', 'else', 'if',
                             'pass', 'yield', 'break', 'except', 'import', 'print',
                             'class', 'exec', 'in', 'raise', 'continue', 'finally',
                             'is', 'return', 'def', 'for', 'lambda', 'try']

        self.allowed_chars=string.digits+string.ascii_letters+'_'

    def Clone(self):
        """ Standard cloner.
            Note that every validator must implement the Clone() method.
        """
        return NoMatchValidTextObjectValidator(self.stringlist)

    def Validate(self, win):
        """ Validate the contents of the given text control.
        """
        textCtrl=self.GetWindow()
        text=textCtrl.GetValue()
        # print text, len(text)
        # print sum([char in self.allowed_chars for char in text])
        if len(text)==0:
            wx.MessageBox("A text object must contain some text!", "Bad Input")
            textCtrl.SetBackgroundColour("pink")
            textCtrl.SetFocus()
            textCtrl.Refresh()
            return False
        elif self.stringlist.__contains__(text):
            wx.MessageBox("Duplicates are not allowed!", "Bad Input")
            textCtrl.SetBackgroundColour("pink")
            textCtrl.SetFocus()
            textCtrl.Refresh()
            return False
        elif text in self.reserved_words:
            wx.MessageBox("Python keywords are not allowed!", "Bad Input")
            textCtrl.SetBackgroundColour("pink")
            textCtrl.SetFocus()
            textCtrl.Refresh()
            return False
        elif sum([char in self.allowed_chars for char in text])!=len(text) \
                or text[0] in string.digits:
            wx.MessageBox("Not a vaild name. Names can only contain letters"
                          ", digits and underscores(_) and not start with a digit.",
                          "Bad Input")
            textCtrl.SetBackgroundColour("pink")
            textCtrl.SetFocus()
            textCtrl.Refresh()
            return False
        else:
            textCtrl.SetBackgroundColour(
                wx.SystemSettings.GetColour(wx.SYS_COLOUR_WINDOW))
            textCtrl.Refresh()
            return True

    def TransferToWindow(self):
        """ Transfer data from validator to window.
            The default implementation returns False, indicating that an error
            occurred.  We simply return True, as we don't do any data transfer.
        """
        return True  # Prevent wxDialog from complaining.

    def TransferFromWindow(self):
        """ Transfer data from window to validator.
            The default implementation returns False, indicating that an error
            occurred.  We simply return True, as we don't do any data transfer.
        """
        return True  # Prevent wxDialog from complaining.

class NoMatchTextCtrlValidator(wx.Validator):
    """ This validator is used to ensure that the user has entered something
        into the text object editor dialog's text field and that it does not match any other string in another textctrl.
    """

    def __init__(self, textctrls):
        """ Standard constructor.
        """
        wx.Validator.__init__(self)
        self.textctrls=textctrls[:]

    def set_nomatch_ctrls(self, textctrls):
        self.textctrls=textctrls[:]

    def Clone(self):
        """ Standard cloner.
            Note that every validator must implement the Clone() method.
        """
        return NoMatchTextCtrlValidator(self.textctrls)

    def Validate(self, win):
        """ Validate the contents of the given text control.
        """
        textCtrl=self.GetWindow()
        text=textCtrl.GetValue()
        stringlist=[ctrl.GetValue() for ctrl in self.textctrls]
        iprint(text, stringlist)
        if len(text)==0:
            wx.MessageBox("A text object must contain some text!", "Error")
            textCtrl.SetBackgroundColour("pink")
            textCtrl.SetFocus()
            textCtrl.Refresh()
            return False
        elif text in stringlist:
            wx.MessageBox("Duplicates are not allowed!", "Error")
            textCtrl.SetBackgroundColour("pink")
            textCtrl.SetFocus()
            textCtrl.Refresh()
            return False
        else:
            textCtrl.SetBackgroundColour(
                wx.SystemSettings.GetColour(wx.SYS_COLOUR_WINDOW))
            textCtrl.Refresh()
            return True

    def TransferToWindow(self):
        """ Transfer data from validator to window.
            The default implementation returns False, indicating that an error
            occurred.  We simply return True, as we don't do any data transfer.
        """
        return True  # Prevent wxDialog from complaining.

    def TransferFromWindow(self):
        """ Transfer data from window to validator.
            The default implementation returns False, indicating that an error
            occurred.  We simply return True, as we don't do any data transfer.
        """
        return True  # Prevent wxDialog from complaining.

class FloatObjectValidator(wx.Validator):
    """ This validator is used to ensure that the user has entered something
        into the text object editor dialog's text field.
    """

    def __init__(self, eval_func=eval, alt_types=None):
        """ Standard constructor.
        """
        wx.Validator.__init__(self)
        self.value=None
        self.eval_func=eval_func
        self.alt_types=alt_types or []

    def Clone(self):
        """ Standard cloner.

            Note that every validator must implement the Clone() method.
        """
        return FloatObjectValidator(self.eval_func, self.alt_types)

    def Validate(self, win):
        """ Validate the contents of the given text control.
        """
        textCtrl=self.GetWindow()
        text=textCtrl.GetValue()
        self.value=None
        try:
            val=self.eval_func(text)
            if is_reflfunction(val):
                val=val.validate()
        except Exception as S:
            wx.MessageBox("Can't evaluate the expression!!\nERROR:\n%s"%S.__str__(), "Error")
            textCtrl.SetBackgroundColour("pink")
            textCtrl.SetFocus()
            textCtrl.Refresh()
            return False
        else:
            try:
                self.value=float(val)
            except Exception as S:
                # print type(val), self.alt_types
                if not any([isinstance(val, typ) for typ in self.alt_types]):
                    wx.MessageBox("Wrong type of parameter\nERROR:\n%s"%S.__str__(), "Error")
                    textCtrl.SetBackgroundColour("pink")
                    textCtrl.SetFocus()
                    textCtrl.Refresh()
                    return False

            textCtrl.SetBackgroundColour(wx.SystemSettings.GetColour(wx.SYS_COLOUR_WINDOW))
            textCtrl.Refresh()
            return True

    def TransferToWindow(self):
        """ Transfer data from validator to window.
            The default implementation returns False, indicating that an error
            occurred.  We simply return True, as we don't do any data transfer.
        """
        return True

    def TransferFromWindow(self):
        """ Transfer data from window to validator.

            The default implementation returns False, indicating that an error
            occurred.  We simply return True, as we don't do any data transfer.
        """
        return True

class ComplexObjectValidator(wx.Validator):
    """ This validator is used to ensure that the user has entered something
       into the text object editor dialog's text field.
    """

    def __init__(self, eval_func=eval, alt_types=None):
        """ Standard constructor.
        """
        wx.Validator.__init__(self)
        self.value=None
        self.eval_func=eval_func
        self.alt_types=alt_types or []

    def Clone(self):
        """ Standard cloner.
            Note that every validator must implement the Clone() method.
        """
        return ComplexObjectValidator(self.eval_func, self.alt_types)

    def Validate(self, win):
        """ Validate the contents of the given text control.
        """
        textCtrl=self.GetWindow()
        text=textCtrl.GetValue()
        self.value=None
        try:
            val=self.eval_func(text)
        except Exception as S:
            wx.MessageBox("Can't compile the complex expression!!\nERROR:\n%s"%S.__str__(), "Error")
            textCtrl.SetBackgroundColour("pink")
            textCtrl.SetFocus()
            textCtrl.Refresh()
            return False
        try:
            # Have to do it differentily to work with proxys
            # self.value=complex(self.eval_func(text))
            if is_reflfunction(val):
                val=val.validate()
            self.value=complex(val.real, val.imag)
        except AttributeError:
            try:
                self.value=complex(val)
            except Exception as S:
                if not any([isinstance(val, typ) for typ in self.alt_types]):
                    wx.MessageBox(
                        "Can't evaluate the complex expression, not the correct type!!\nERROR:\n%s"%S.__str__(),
                        "Error")
                    textCtrl.SetBackgroundColour("pink")
                    textCtrl.SetFocus()
                    textCtrl.Refresh()
                    return False
                else:
                    textCtrl.SetBackgroundColour(wx.SystemSettings.GetColour(wx.SYS_COLOUR_WINDOW))
                    textCtrl.Refresh()
                    return True
        except Exception as S:
            if not any([isinstance(val, typ) for typ in self.alt_types]):
                wx.MessageBox("Can't evaluate the complex expression!!\nERROR:\n%s"%S.__str__(), "Error")
                textCtrl.SetBackgroundColour("pink")
                textCtrl.SetFocus()
                textCtrl.Refresh()
                return False
            else:
                textCtrl.SetBackgroundColour(wx.SystemSettings.GetColour(wx.SYS_COLOUR_WINDOW))
                textCtrl.Refresh()
                return True
        else:
            # print 'OK'
            textCtrl.SetBackgroundColour(
                wx.SystemSettings.GetColour(wx.SYS_COLOUR_WINDOW))
            textCtrl.Refresh()
            return True

    def TransferToWindow(self):
        """ Transfer data from validator to window.
            The default implementation returns False, indicating that an error
            occurred.  We simply return True, as we don't do any data transfer.
        """
        return True

    def TransferFromWindow(self):
        """ Transfer data from window to validator.
            The default implementation returns False, indicating that an error
            occurred.  We simply return True, as we don't do any data transfer.
        """
        return True

# ----------------------------------------------------------------------


class ValidateBaseDialog(wx.Dialog):
    def __init__(self, parent, pars, vals, validators, title="Validated Dialog",
                 units=None, groups=None, cols=2, editable_pars=None):
        ''' A dialog that validates the input when OK are pressed. The validation is done
        through the validators given as input.
        
        Pars should contain a list of names.
        Vals should be a dictionary of names and values for the different parameters one wish to set. 
        editable_pars should be a dictonary indicating if the parameter should be editable, a missing value
        is interpreted that the parameter is editable.
        Groups should either be false or consist of a list of lists which
        in turn consist of the (name, value) tuples where the first item
        in the list should be a string describing the group. This will be layed out 
        with subboxes. An example: 
        groups = [['Standard', [('f', 25), ('sigma', 7)]], ['Neutron', [('b', 3.0)]]]


        Note validators, values, units and editable pars should be dictionaries of values!
        '''
        wx.Dialog.__init__(self, parent, -1, title)
        self.pars=pars
        self.editable_pars=editable_pars or {}
        self.validators=validators
        self.cols=cols
        self.vals=vals
        self.units=units or {}
        self.groups=groups or []
        self.tc={}
        self.not_editable_bkg_color=wx.GREEN
        self.SetAutoLayout(True)

        if self.groups:
            self.main_sizer, self.tc=self.grid_layout(self, self.vals, self.editable_pars)
        else:
            self.main_sizer, self.tc=self.layout_group(self, self.pars, self.vals, self.editable_pars)

        self.main_layout()
        self.Layout()

    def main_layout(self):
        border=wx.BoxSizer(wx.VERTICAL)
        border.Add(self.main_sizer, 1, wx.GROW | wx.ALL, 5)

        line=wx.StaticLine(self, -1, size=(20, -1), style=wx.LI_HORIZONTAL)
        border.Add(line, 0, wx.GROW | wx.RIGHT | wx.TOP, 5)

        buttons=self.create_buttons()
        border.Add(buttons, flag=wx.ALIGN_RIGHT | wx.ALL, border=5)
        self.SetSizer(border)
        border.Fit(self)
        self.border_sizer=border

    def create_buttons(self):
        buttons=wx.StdDialogButtonSizer()  # wx.BoxSizer(wx.HORIZONTAL)
        b=wx.Button(self, wx.ID_OK, "OK")
        b.SetDefault()
        buttons.AddButton(b)
        buttons.AddButton(wx.Button(self, wx.ID_CANCEL, "Cancel"))
        buttons.Realize()
        return buttons

    def grid_layout(self, parent, vals, editable_pars):
        '''Do an more advanced layout with subboxes
        '''
        rows=math.ceil(len(self.pars)/(self.cols*1.0))
        sizer=wx.FlexGridSizer(rows=rows, cols=self.cols,
                               vgap=10, hgap=10)
        tc={}
        for group in self.groups:
            if type(group[0])!=str:
                raise TypeError('First item in a group has to be a string')
            # Make the box for putting in the columns
            col_box=wx.StaticBox(parent, -1, group[0])
            col_box_sizer=wx.StaticBoxSizer(col_box, wx.VERTICAL)
            group, group_tc=self.layout_group(parent, group[1], vals, editable_pars)
            col_box_sizer.Add(group, flag=wx.EXPAND, border=5)
            for item in group_tc:
                tc[item]=group_tc[item]
            sizer.Add(col_box_sizer, flag=wx.EXPAND)
        return sizer, tc

    def layout_group(self, parent, pars, vals, editable_pars):
        if self.units:
            layout_cols=3
        else:
            layout_cols=2
        sizer=wx.FlexGridSizer(len(pars)+1, layout_cols,
                               vgap=10, hgap=5)
        tc={}
        for par in pars:
            # print par, vals[par]
            label=wx.StaticText(parent, -1, par+': ')
            validator=self.validators[par]  # .Clone()
            val=vals[par]
            editable=True
            if par in editable_pars:
                editable=editable_pars[par]

            tc[par]=self.crete_edit_ctrl(editable, par, parent, val, validator)

            sizer.Add(label,
                      flag=wx.ALIGN_RIGHT | wx.ALIGN_CENTER_VERTICAL, border=5)
            sizer.Add(tc[par],
                      flag=wx.ALIGN_CENTER_VERTICAL | wx.EXPAND, border=5)

            # If we include units as well:
            if self.units:
                unit_label=wx.StaticText(parent, -1, ' '+self.units[par])
                sizer.Add(unit_label,
                          flag=wx.ALIGN_LEFT | wx.ALIGN_CENTER_VERTICAL,
                          border=5)
        return sizer, tc

    def GetValues(self):
        # print dir(self.tc[0])
        # print self.tc[0].GetValue()
        p={}
        for par in list(self.validators.keys()):
            if type(self.validators[par])==type([]):
                # have to pad teh text to make it a string inside a string...
                # text = '\''
                text=self.validators[par][self.tc[par].GetSelection()]
                # text += '\''
                p[par]=text
            else:
                p[par]=self.tc[par].GetValue()
        return p

class NormalEditMixin:
    """
    Normal mixin for the Validate Dialog
    """
    not_editable_bkg_color=None

    def crete_edit_ctrl(self, editable, par, parent, val, validator):
        """
        Creates the edit control for a parameter
        :param editable: flag to indicate if the parameter is editable
        :param par: parameter name
        :param parent: parent of the control
        :param val: value of the parameter
        :param validator: validator of the control
        :return: a wx control object (subclass of wx.Control)?
        """
        if type(validator)==type([]):
            # There should be a list of choices
            validator=validator[:]
            ctrl=wx.Choice(parent, -1, choices=validator)
            # Since we work with strings we have to find the right
            # strings positons to initilize the choice box.
            pos=0
            if type(val)==type(''):
                pos=validator.index(val)
            elif type(par)==type(1):
                pos=par
            ctrl.SetSelection(pos)
        elif isinstance(validator, bool):
            # Parameter is a boolean
            ctrl=wx.CheckBox(self, -1)
            ctrl.SetValue(val)
            # Create a non-editable box if needed
            # ctrl.SetEditable(editable)
        elif isinstance(validator, int):
            # Parameter is an integer
            ctrl=IntCtrl(self, -1, val)
            # Create a non-editable box if needed
            ctrl.SetEditable(editable)
            if not editable:
                ctrl.SetBackgroundColour(self.not_editable_bkg_color)

        # Otherwise it should be a validator ...
        else:
            validator=validator.Clone()
            ctrl=wx.TextCtrl(parent, -1, str(val),
                             validator=validator,
                             style=wx.TE_RIGHT | wx.TE_RICH)

            # Create a non-editable box if needed
            ctrl.SetEditable(editable)
            if not editable:
                ctrl.SetBackgroundColour(self.not_editable_bkg_color)
        return ctrl

class ValidateBaseNotebookDialog(ValidateBaseDialog):
    def __init__(self, parent, pars, vals, validators, title="Validated Dialog",
                 units=None, groups=None, cols=2, fixed_pages=None,
                 editable_pars=None):
        ''' A dialog that validates the input when OK are pressed. The validation is done
        through the validators given as input.
        
        Pars should contain a list of names.
        Vals should be a dictionary of dictionaries of names and values for the different parameters one wish to set.
        editable_pars should be a dictionary of the same form as vals.
        The upper dictionary level should contain the object names.
        Groups should either be false or a consist of a list of lists which
        in turn consist of the (name, value) tuples where the first item
        in the list should be a string describing the group. This will be layed out 
        with subboxes. An example: 
        groups = [['Standard', ('f', 25), ('sigma', 7)], ['Neutron', ('b', 3.0)]]
        
        Note validators, values and units should be dictonaries of values!
        '''
        wx.Dialog.__init__(self, parent, -1, title)
        self.SetExtraStyle(wx.WS_EX_VALIDATE_RECURSIVELY)
        self.pars=pars
        self.validators=validators
        self.cols=cols
        self.vals=vals
        self.units=units or {}
        self.groups=groups or []
        self.fixed_pages=fixed_pages or []
        self.changes=[]
        self.text_controls={}
        self.SetAutoLayout(True)
        self.editable_pars=editable_pars or {}
        self.not_editable_bkg_color=wx.GREEN

        self.layout_notebook()
        self.do_toolbar()

        self.main_layout()
        self.Layout()

    def main_layout(self):
        border=wx.BoxSizer(wx.VERTICAL)
        border.Add((-1, 5))
        border.Add(self.toolbar)
        # border.Add((-1,2))

        border.Add(self.main_sizer, 1, wx.GROW | wx.ALL, 5)

        line=wx.StaticLine(self, -1, size=(20, -1), style=wx.LI_HORIZONTAL)
        border.Add(line, 0, wx.GROW | wx.RIGHT | wx.TOP, 5)

        buttons=self.create_buttons()
        border.Add(buttons, flag=wx.ALIGN_RIGHT | wx.ALL, border=5)
        self.SetSizer(border)
        border.Fit(self)
        self.border_sizer=border

    def do_toolbar(self):
        if os.name=='nt':
            size=(24, 24)
        else:
            size=(-1, -1)
        dpi_scale_factor=wx.GetApp().dpi_scale_factor
        tb_bmp_size=int(dpi_scale_factor*20)

        space=(5, -1)
        button_names=['Insert', 'Delete', 'Rename']
        button_images=[wx.Bitmap(images.insert_layer.GetImage().Scale(tb_bmp_size, tb_bmp_size)),
                       wx.Bitmap(images.delete.GetImage().Scale(tb_bmp_size, tb_bmp_size)),
                       wx.Bitmap(images.change_name.GetImage().Scale(tb_bmp_size, tb_bmp_size))]
        callbacks=[self.eh_insert, self.eh_delete, self.eh_rename]
        tooltips=['Insert', 'Delete', 'Rename']

        boxbuttons=wx.BoxSizer(wx.HORIZONTAL)
        boxbuttons.Add((5, -1))
        for i in range(len(button_names)):
            # button = wx.Button(self,-1, button_names[i])
            button=wx.BitmapButton(self, -1, button_images[i],
                                   style=wx.NO_BORDER, size=size)
            boxbuttons.Add(button, 1, wx.EXPAND, 5)
            boxbuttons.Add(space)
            button.SetToolTip(tooltips[i])
            self.Bind(wx.EVT_BUTTON, callbacks[i], button)
        self.toolbar=boxbuttons

    def layout_notebook(self):
        '''Do the notebook layout in the dialog
        '''
        self.main_sizer=wx.Notebook(self, -1, style=wx.NB_TOP)
        names=list(self.vals.keys())
        names.sort()
        for item in names:
            editable_pars={}
            if item in self.editable_pars:
                editable_pars=self.editable_pars[item]
            self.AddPage(item, self.vals[item], editable_pars)

    def eh_insert(self, evt):
        '''Eventhandler that creates a new panel
        '''
        pos=self.main_sizer.GetSelection()
        current_name=self.main_sizer.GetPageText(pos)
        if len(self.fixed_pages)>0:
            current_name=self.fixed_pages[0]
        index=1
        while '%s_%d'%(current_name, index) in list(self.vals.keys()):
            index+=1
        new_name='%s_%d'%(current_name, index)

        state={}
        for key in self.vals[current_name]:
            state[key]=0

        self.AddPage(new_name, self.vals[current_name], state, select=True)
        self.changes.append(('', new_name))

    def eh_delete(self, evt):
        '''Eventhandler that deletes the current page and its data
        '''
        pos=self.main_sizer.GetSelection()
        current_name=self.main_sizer.GetPageText(pos)
        self.RemovePage(current_name)
        self.changes.append((current_name, ''))

    def eh_rename(self, evt):
        '''Eventhandler that renames the current tab
        '''
        pos=self.main_sizer.GetSelection()
        current_name=self.main_sizer.GetPageText(pos)
        if current_name in self.fixed_pages:
            wx.MessageBox('It is forbidden to change the' \
                          'name of %s.'%current_name)
        else:
            unallowed_names=[]
            for name in self.vals:
                if name!=current_name:
                    unallowed_names.append(name)
            dlg=ValidateDialog(self, ['Name'], {'Name': current_name},
                               {'Name': NoMatchValidTextObjectValidator(unallowed_names)},
                               title='Give New Name')

            if dlg.ShowModal()==wx.ID_OK:
                vals=dlg.GetValues()
                self.vals[vals['Name']]=self.vals.pop(current_name)
                self.text_controls[vals['Name']]=self.text_controls.pop(current_name)
                self.main_sizer.SetPageText(pos, vals['Name'])
                self.changes.append((current_name, vals['Name']))
            dlg.Destroy()

    def AddPage(self, name, vals, editable_pars, select=False):
        '''Adds a page to the notebook 
        
        the flag select activates a page change to the current page
        '''
        panel=wx.Panel(self.main_sizer, -1)
        self.main_sizer.AddPage(panel, name, select=select)
        if self.groups:
            sizer, page_text_controls=self.grid_layout(panel, vals, editable_pars)
        else:
            sizer, page_text_controls=self.layout_group(panel, self.pars, vals, editable_pars)
        self.text_controls[name]=page_text_controls
        panel.SetSizerAndFit(sizer)
        panel.Layout()
        self.Fit()
        # Make sure that the data is added if we add a new page after init has
        # been executed
        if not name in self.vals:
            self.vals[name]=vals.copy()
            self.editable_pars[name]=editable_pars.copy()

    def RemovePage(self, name):
        '''Remove a page from the notebook
        '''
        sucess=False
        index=-1
        if name in self.fixed_pages:
            return sucess
        for i in range(self.main_sizer.GetPageCount()):
            if self.main_sizer.GetPageText(i)==name:
                sucess=True
                index=i
        if sucess:
            self.main_sizer.DeletePage(index)
            self.vals.pop(name)
        return sucess

    def GetValues(self):
        '''Extracts values from the text controls and returns is as a 
        two level nested dictionary.
        '''
        p={}
        for page in self.vals:
            p[page]={}
            for par in list(self.validators.keys()):
                if type(self.validators[par])==type([]):
                    # have to pad teh text to make it a string inside a string...
                    # text = '\''
                    text=self.validators[par][self.text_controls[page][par].GetSelection()]
                    # text += '\''
                    p[page][par]=text
                else:
                    p[page][par]=self.text_controls[page][par].GetValue()
        return p

    def GetChanges(self):
        '''GetChanges(self) --> changes
        Returns the changes as a list consisting of tuples (old_name, new_name).
        if the page has been deleted new_name is an empty string.
        '''
        return self.changes

class FitEditMixIn:
    """
    Mixin for the Validate Dialogs to enable the definition of fitable parameters.
    """
    validators=None
    tc=None

    def __init__(self):
        pass

    def crete_edit_ctrl(self, state, par, parent, val, validator):
        """
        Creates the edit control for a parameter
        :param state: flag to indicate if the parameter is editable integer that takes the values
            0 - editable
            1 - fitted in the grid
            2 - constant but defined in the grid
        :param par: parameter name
        :param parent: parent of the control
        :param val: value of the parameter
        :param validator: validator of the control
        :return: a wx control object (subclass of wx.Control)?
        """
        if type(validator)==type([]):
            # There should be a list of choices
            validator=validator[:]
            ctrl=wx.Choice(parent, -1,
                           choices=validator)
            # Since we work with strings we have to find the right
            # strings positons to initilize the choice box.
            pos=0
            if type(val)==type(''):
                pos=validator.index(val)
            elif type(par)==type(1):
                pos=par
            ctrl.SetSelection(pos)
        # Otherwise it should be a validator ...
        else:
            validator=validator.Clone()
            ctrl=FitSelectorCombo(state, parent, -1, str(val),
                                  validator=validator,
                                  style=wx.TE_RICH | wx.TE_RIGHT)
        return ctrl

    def GetStates(self):
        """ Returns the states of the parameters in the dialog

        :return: a dictionary with keys representing parameters and values representing state.
        """
        # print dir(self.tc[0])
        # print self.tc[0].GetValue()
        p={}
        for par in list(self.validators.keys()):
            if isinstance(self.validators[par], list):
                p[par]=0
            else:
                p[par]=self.tc[par].state
        return p

class ValidateDialog(ValidateBaseDialog, NormalEditMixin):
    """
    The normal Validate Dialog definition
    """

class ValidateFitDialog(ValidateBaseDialog, FitEditMixIn):
    """
    A sub-class of ValidateDialog that uses a custom ComboCtrl to allow a parameter to be
    set for fitting.
    """

class ValidateNotebookDialog(ValidateBaseNotebookDialog, NormalEditMixin):
    """
    The normal Validate Notebook Dialog.
    """

class ValidateFitNotebookDialog(ValidateBaseNotebookDialog, FitEditMixIn):
    """
    A sub-class of ValidateNotebookDialog that uses a custom ComboCtrl to allow a parameter to be
    set for fitting.
    """

    def GetStates(self):
        '''Extracts values from the text controls and returns is as a
        two level nested dictionary.
        '''
        p={}
        for page in self.vals:
            p[page]={}
            for par in list(self.validators.keys()):
                if isinstance(self.validators[par], list):
                    p[page][par]=0
                else:
                    p[page][par]=self.text_controls[page][par].state
        return p

class FitSelectorCombo(wx.ComboCtrl):
    """Class that defines a Combobox that allows the user to set if the parameter should be handled by the grid
    , fitted or constants from the beginning.
    """

    def update_text_state(self):
        self.SetEditable(not self.state)
        # self.GetTextCtrl().Enable(not self.state)
        font=self.GetFont()

        if self.state==0:
            self.SetForegroundColour(wx.BLACK)
            font.SetWeight(wx.FONTWEIGHT_NORMAL)
        elif self.state==1:
            self.SetForegroundColour(self.fit_bkg_color)
            font.SetWeight(wx.FONTWEIGHT_BOLD)
        elif self.state==2:
            self.SetForegroundColour(self.const_fit_bkg_color)
            font.SetWeight(wx.FONTWEIGHT_BOLD)
        self.SetFont(font)

        self.Refresh()

    def create_button_bitmap(self):
        # make a custom bitmap showing "F"
        bw, bh=14, 16
        bmp=wx.Bitmap(bw, bh, depth=wx.BITMAP_SCREEN_DEPTH)
        dc=wx.MemoryDC(bmp)
        # clear to a specific background colour
        bgcolor=wx.Colour(255, 254, 255)
        dc.SetBackground(wx.Brush(bgcolor))
        dc.Clear()
        # draw the label onto the bitmap
        label="F"
        font=wx.SystemSettings.GetFont(wx.SYS_DEFAULT_GUI_FONT)
        font.SetWeight(wx.FONTWEIGHT_BOLD)
        dc.SetFont(font)
        tw, th=dc.GetTextExtent(label)
        dc.DrawText(label, (bw-tw)/2, (bh-th)/2)
        del dc
        # now apply a mask using the bgcolor
        bmp.SetMaskColour(bgcolor)
        # and tell the ComboCtrl to use it
        self.SetButtonBitmaps(bmp, True)

    def __init__(self, state, *args, **kw):
        """

        :param state: an int signalling the state 0 - parameter defined in dialog,
            1 - parameter fitted, 2 - parameter constant in grid.
        :param args:
        :param kw:
        :return:
        """
        wx.ComboCtrl.__init__(self, *args, **kw)

        self.state=state
        # Green wx.Colour(138, 226, 52), ORANGE wx.Colour(245, 121, 0)
        self.fit_bkg_color=wx.Colour(245, 121, 0)
        # Tango Sky blue wx.Colour(52, 101, 164), wx.Colour(114, 159, 207)
        self.const_fit_bkg_color=wx.Colour(114, 159, 207)

        self.create_button_bitmap()
        self.update_text_state()

        # Set the id's for the Menu
        self.id_define_par=wx.NewId()
        self.id_fit_par=wx.NewId()
        self.id_const_fit_par=wx.NewId()

        menu=wx.Menu()
        menu.AppendRadioItem(self.id_define_par, 'Define parameter here',
                             'Unlock the parameter for editing and remove it from the grid.')
        menu.AppendRadioItem(self.id_fit_par, 'Fit parameter',
                             'Define the parameter in the grid and use fit it.')
        menu.AppendRadioItem(self.id_const_fit_par, 'Constant fit parameter',
                             'Define the parameter in the fri')
        if not self.state:
            menu.Check(self.id_define_par, True)
        elif self.state==1:
            menu.Check(self.id_fit_par, True)
        elif self.state==2:
            menu.Check(self.id_const_fit_par, True)

        self.menu=menu

    # Overridden from ComboCtrl, called when the combo button is clicked
    def OnButtonClick(self):
        # Create a popup menu when pressing the button.
        self.PopupMenu(self.menu)
        if self.menu.IsChecked(self.id_define_par):
            self.state=0
        elif self.menu.IsChecked(self.id_fit_par):
            self.state=1
        elif self.menu.IsChecked(self.id_const_fit_par):
            self.state=2
        self.update_text_state()

    # Overridden from ComboCtrl to avoid assert since there is no ComboPopup
    def DoSetPopupControl(self, popup):
        pass

class ParameterExpressionCombo(wx.ComboCtrl):
    """Class that defines a Combobox for editing an expression and inserting parameters from the model in that
    expression by choosing them from a popupmenu.
    """

    def create_button_bitmap(self):
        # make a custom bitmap showing "F"
        bw, bh=14, 16
        bmp=wx.Bitmap(bw, bh, depth=wx.BITMAP_SCREEN_DEPTH)
        dc=wx.MemoryDC(bmp)
        # clear to a specific background colour
        bgcolor=wx.Colour(255, 254, 255)
        dc.SetBackground(wx.Brush(bgcolor))
        dc.Clear()
        # draw the label onto the bitmap
        label="P"
        font=wx.SystemSettings.GetFont(wx.SYS_DEFAULT_GUI_FONT)
        font.SetWeight(wx.FONTWEIGHT_BOLD)
        dc.SetFont(font)
        tw, th=dc.GetTextExtent(label)
        dc.DrawText(label, (bw-tw)/2, (bh-th)/2)
        del dc
        # now apply a mask using the bgcolor
        bmp.SetMaskColour(bgcolor)
        # and tell the ComboCtrl to use it
        self.SetButtonBitmaps(bmp, True)

    def __init__(self, par_dict, sim_func, *args, **kw):
        """

        :param state: an int signalling the state 0 - parameter defined in dialog,
            1 - parameter fitted, 2 - parameter constant in grid.
        :param args:
        :param kw:
        :return:
        """
        wx.ComboCtrl.__init__(self, *args, **kw)

        self.create_button_bitmap()

        # Set the id's for the Menu
        self.id_simulate=wx.NewId()

        self.par_dict=par_dict
        self.sim_func=sim_func

    # Overridden from ComboCtrl, called when the combo button is clicked
    def OnButtonClick(self):
        menu=wx.Menu()
        par_dict=self.par_dict
        classes=list(par_dict.keys())
        classes.sort(key=str.lower)
        for cl in classes:
            # Create a submenu for each class
            clmenu=wx.Menu()
            obj_dict=par_dict[cl]
            objs=list(obj_dict.keys())
            objs.sort(key=str.lower)
            # Create a submenu for each object
            for obj in objs:
                obj_menu=wx.Menu()
                funcs=obj_dict[obj]
                funcs.sort(key=str.lower)
                # Create an item for each method
                for func in funcs:
                    item=obj_menu.Append(wx.NewId(), obj+'.'+func.replace('set', 'get'))
                    self.Bind(wx.EVT_MENU, self.OnPopUpItemSelected, item)
                clmenu.Append(-1, obj, obj_menu)
            menu.Append(-1, cl, clmenu)
        # Check if there are no available classes
        # TODO: Test implementation if necessary?
        # if len(classes) == 0:
        #    # Add an item to compile the model
        #    item = menu.Append(self.id_simulate, 'Simulate to see parameters')
        #    self.Bind(wx.EVT_MENU, self.OnSimulate, item)
        # Create a popup menu when pressing the button.
        self.menu=menu
        self.PopupMenu(self.menu)
        menu.Destroy()

    def OnSimulate(self, event):
        """ Event handler for selection of the simulate option in the popup.
        :param event:
        :return:
        """
        if not self.sim_func is None:
            self.sim_func()

    def OnPopUpItemSelected(self, event):
        """ Event handler for selcting a parameters get function.

        :param event:
        :return:
        """
        item=self.menu.FindItemById(event.GetId())
        self.WriteText(item.GetItemLabel()+'()')

    # Overridden from ComboCtrl to avoid assert since there is no ComboPopup
    def DoSetPopupControl(self, popup):
        pass

class ZoomFrame(wx.MiniFrame):
    def __init__(self, parent):
        wx.MiniFrame.__init__(self, parent, -1, "X-Y Scales")

        # self.SetAutoLayout(True)

        VSPACE=10

        self.panel=wx.Panel(self, -1, style=wx.TAB_TRAVERSAL
                                            | wx.CLIP_CHILDREN
                                            | wx.FULL_REPAINT_ON_RESIZE
                            )

        gbs=wx.GridBagSizer(3, 3)
        label=wx.StaticText(self.panel, -1, 'Min')
        gbs.Add(label, (0, 1), flag=wx.ALIGN_CENTER, border=2)
        label=wx.StaticText(self.panel, -1, 'Max')
        gbs.Add(label, (0, 2), flag=wx.ALIGN_CENTER, border=2)
        label=wx.StaticText(self.panel, -1, ' X'+': ')
        gbs.Add(label, (1, 0), flag=wx.ALIGN_RIGHT, border=2)
        self.xmin=wx.TextCtrl(self.panel, -1, '0', validator=FloatObjectValidator())
        gbs.Add(self.xmin, (1, 1), flag=wx.ALIGN_CENTER | wx.EXPAND, border=2)
        self.xmax=wx.TextCtrl(self.panel, -1, '0', validator=FloatObjectValidator())
        gbs.Add(self.xmax, (1, 2), flag=wx.ALIGN_CENTER | wx.EXPAND, border=2)
        label=wx.StaticText(self.panel, -1, ' Y'+': ')
        gbs.Add(label, (2, 0), flag=wx.ALIGN_RIGHT, border=2)
        self.ymin=wx.TextCtrl(self.panel, -1, '0', validator=FloatObjectValidator())
        gbs.Add(self.ymin, (2, 1), flag=wx.ALIGN_CENTER | wx.EXPAND, border=2)
        self.ymax=wx.TextCtrl(self.panel, -1, '0', validator=FloatObjectValidator())
        gbs.Add(self.ymax, (2, 2), flag=wx.ALIGN_CENTER | wx.EXPAND, border=2)

        buttons=wx.BoxSizer(wx.HORIZONTAL)
        b=wx.Button(self.panel, wx.ID_OK, "Apply")
        # b.SetDefault()
        # buttons.AddButton(wx.Button(self, wx.ID_CANCEL, "Cancel"))
        # buttons.Realize()
        buttons.Add((0, 0), 2, wx.EXPAND)
        buttons.Add(b, 1, flag=wx.ALIGN_RIGHT)
        border=wx.BoxSizer(wx.VERTICAL)

        border.Add(gbs, 0, wx.GROW | wx.ALL, 2)
        border.Add(buttons)

        self.panel.SetSizerAndFit(border)
        self.SetClientSize(self.panel.GetSize())
        # self.Layout()

if __name__=='__main__':
    class MyApp(wx.App):
        def OnInit(self):
            # wx.InitAllImageHandlers()
            frame=ZoomFrame(None)
            # frame.Show(True)
            # self.SetTopWindow(frame)
            pars=['test1', 'test2']
            vals={'inst': {'test1': 3, 'test2': 5.0J}, 'x-ray': {'test1': 3, 'test2': 5.0+9J}}
            vals_old={'test1': 3, 'test2': 7.0J}
            validators={'test1': FloatObjectValidator(), 'test2': ComplexObjectValidator()}
            units={'test1': 'barn', 'test2': 'm/s'}
            dlg=ValidateNotebookDialog(frame, pars, vals, validators,
                                       title='Instrument Editor', units=units,
                                       fixed_pages=['inst'])
            dlg.AddPage('test', vals_old, {})
            dlg.RemovePage('inst')
            # dlg = ValidateDialog(frame, pars, vals_old, validators,
            #                 title = 'Instrument Editor', units = units)
            if dlg.ShowModal()==wx.ID_OK:
                # print 'Pressed OK'
                vals=dlg.GetValues()
                iprint(vals)
            return True

    app=MyApp(0)
    # app.MainLoop()




from bumps.monitor import TimedUpdate
from bumps.fitproblem import nllf_scale
from bumps.formatnum import format_uncertainty

class NBMonitor(TimedUpdate):
    """
    Display fit progress on the console
    """

    def __init__(self, problem, progress=0.25, improvement=5.0):
        TimedUpdate.__init__(self, progress=progress, improvement=improvement)
        from ipywidgets import Output, HTML, HBox, VBox
        from IPython.display import display
        self.problem=problem
        self.status_text=HTML()
        self.result_text=HTML()
        self.plot_out=Output()

        vbox=VBox([self.status_text, self.result_text])
        hbox=HBox([vbox, self.plot_out])
        display(hbox)
        self.steps=[]
        self.chis=[]

    def show_progress(self, history):
        scale, err=nllf_scale(self.problem)
        chisq=format_uncertainty(scale*history.value[0], err)
        self.status_text.value='<table width="50%%"><tr><td>step</td><td>%s</td><td>cost</td><td>%s</td></tr></table>'%(
        history.step[0], chisq)
        self.steps.append(history.step[0])
        self.chis.append(scale*history.value[0])

    def show_improvement(self, history):
        from IPython.display import display, clear_output
        from matplotlib.pyplot import figure, plot, close
        p=self.problem.getp()
        try:
            self.problem.setp(history.point[0])
            out='<table width="50%"><tr>'
            out+="</tr>\n<tr>".join(
                ["<td>%s</td><td>%s</td><td>%s</td>"%(pi.name, pi.value, pi.bounds) for pi in self.problem._parameters])
            self.result_text.value=out+"</td></tr></table>"
        finally:
            self.problem.setp(p)
        with self.plot_out:
            fig=figure()
            plot(self.steps, self.chis)
            clear_output()
            display(fig)
            close()




r"""
.. module:: parameters
   :synopsis: Classes to work with differnt materials in scattering experiments

"""

from os import path
import re
import pickle

import numpy as np
from scipy import interpolate

from .parameters import HasParameters, ComplexArray, Float, Calc
from .physical_constants import r_e
from . import scatteringlengths as _sl

# Transformations between AA and eV
_AA_to_eV=12398.42

_module_dir, _filename=path.split(__file__)
# Look only after the file name and not the ending since
# the file ending can be pyc if compiled...
_filename=_filename.split('.')[0]
_module_dir='.' if _module_dir=='' else _module_dir

_atomic_weights=_sl.load_atomic_weights_dabax(path.join(_module_dir, "../databases/AtomicWeights.dat"))

def create_fpdisp_lookup(path):
    """ Creates a lookup function to lookup element names and returns a function
    that yields a function of dispersive scattering factors f(E) at Q = 0. NOTE energy is in eV
    """

    def create_dispersion_func(name):
        e, f1, f2=np.loadtxt(path+'%s.nff'%name.lower(), skiprows=1, unpack=True)
        f1interp=interpolate.interp1d(e, f1, kind='linear')
        f2interp=interpolate.interp1d(e, f2, kind='linear')
        def_wl=_AA_to_eV/np.mean(e)

        @MemoizeF
        def f(wl=def_wl, **kwargs):
            return f1interp(_AA_to_eV/wl)-1.0J*f2interp(_AA_to_eV/wl)

        return Calc(f)

    return create_dispersion_func

f0=create_fpdisp_lookup(path.join(_module_dir, "../databases/f1f2_nist/"))

class MemoizeF:
    """Remember previous evaluations of function fn and stores it in a dictionary.

    Based on the Cookbook recipe: Memoizing (Caching) the Return Values of Functions in
    Python Cookbook by David Ascher, Alex Martelli.
    """

    def __init__(self, fn):
        self.fn=fn
        self.memo={}

    def __call__(self, wl=1.54, **kwds):
        pickled=pickle.dumps(wl, 1)
        if pickled not in self.memo:
            self.memo[pickled]=self.fn(wl=wl, **kwds)
        return self.memo[pickled]

class Material(HasParameters):
    """ Class to keep transform a chemical formula and mass density to scattering length

    The formula is given in ASCII format, for example 'Ag2(NO2)3' or 'HNO3', nested parentheses are allowed.
    Also non-integer indices are allowed for example Fe0.01Pd

    """
    validation_kwargs={'wl': 1.54}

    # The scattering length should be in AA-2
    # Need to transform g/cm3 to formula units per AA3
    # 1u = 1.66054×10−27 kg 1 AA=10−10m
    # density[kg/m3] = 1.66054*10^3*density[u/A3]
    # density [g/cm3] = density [kg/m3] * 1e-3/1e-2**3 = density [kg/m3] * 1e3
    # So formula units per AA^ = density [kg/m3]/1.66054*10^3/weight per formula unit
    # or formula units per AA^ = density [g/cm3]/1.66054/weight per formula unit

    def __init__(self, formula, **kwargs):
        self.sld_x=ComplexArray(0, unit="1/AA-2", help="Scattering length density")
        self.density=Float(0.0, unit="g/cm3", help="Weight density")
        self._formula_density=Float(0.0, unit="u", help="weight of one formula unit")
        HasParameters.__init__(self, **kwargs)
        self.formula_dict={}
        self.parse_formula(formula)

        # Create the composition variables needed to describe formula
        w_list=[]
        f0_list=[]
        for name in self.formula_dict:
            self.__setattr__(name, Float(self.formula_dict[name], help="Number of %s atoms"%name))
            element_count=self.__getattribute__(name)
            w_list.append(element_count*self.get_atomic_weight(name))
            f0_list.append(element_count*self.get_f0(name))
        self._formula_density=self.density/1.66054/sum(w_list)
        self.sld_x=self._formula_density*sum(f0_list)*r_e

    def get_atomic_weight(self, element):
        """Returns the atomic weight in unified atomic mass units (u)

        Parameters:
            element(string): name of the element

        """
        return _atomic_weights[element.lower()]

    def get_f0(self, element):
        """Returns the complex atomic forward scattering length normalised to the scattering length of one electron for
         element.

        Parameters:
            element(string): name of the element

        Returns:
            f0(function): function for f0 that takes the wavelength as a keyword argument
        """
        return f0(element)

    def parse_formula(self, formula, multiplier=1):
        """Parses the string formula and returns a list of isotopes

        Parameters:
            formula(string): A string containing the formula.
            multiplier(int): A number representing that will be used to multiply the element count
        """
        # regexp = re.compile(r"([A-Z][a-z]?)(\d*)|\((.*)\)(\d*)")
        regexp=re.compile(r"([A-Z][a-z]?)(\d*\.?\d*)|\((.*)\)(\d*\.?\d*)")

        element, element_count, group, group_count=(1, 2, 3, 4)
        pos=0
        while pos<len(formula):
            match=regexp.match(formula, pos)
            if not match:
                raise ValueError("Error in formula: %s at position %d"%(formula, pos))
            pos=match.end()
            if match.group(element):
                # print "Match %s is an element" % match.group(0)
                count=self.formula_count(match.group(element_count), multiplier)
                try:
                    self.formula_dict[match.group(element)]+=count
                except KeyError:
                    self.formula_dict[match.group(element)]=count
            elif match.group(group):
                # print "Group found: %s, count: %s" % (match.group(group), match.group(group_count))
                count=self.formula_count(match.group(group_count), multiplier)
                self.parse_formula(match.group(group), multiplier=count)

    def formula_count(self, count_str, multiplier):
        """Handles the parsing result for the count of an formula part. Returns the count*multiplier"

        Parameters:
            count_str(string): The count of the element/group as a string
            multiplier(float): The  multiplier from previous groups.

        """
        if not count_str:
            # Set count to 1 if count is not exist, ie a single group is parsed
            count=1
        else:
            count=float(count_str)
        return count*multiplier



''' <h1>Simple Reflectivity plugin </h1>
Reflectivity is a plugin for beginners just showing one single tab for
the sample and hiding all complex GenX functionality<p>

The plugin consists of the following components:
<h2>Sample tab</h2>
This tab has the definiton for the layers in a table.

<h2>SLD tab</h2>
This shows the real and imaginary part of the scattering length as a function
of depth for the sample. The substrate is to the left and the ambient material
is to the right. This is updated when the simulation button is pressed.
'''

from .. import add_on_framework as framework
from genx.exceptions import GenxError
from genx.gui.solvergui import EVT_UPDATE_PARAMETERS
from genx.gui.parametergrid import EVT_PARAMETER_GRID_CHANGE
import wx.grid as gridlib
from wx.adv import Wizard, WizardPageSimple

import io, traceback

from .Reflectivity import SamplePlotPanel, find_code_segment
from .help_modules.custom_dialog import *
from .help_modules import reflectivity_images as images
from genx.gui.images import getopenBitmap, getplottingBitmap
from .help_modules.materials_db import mdb, Formula, MASS_DENSITY_CONVERSION
from genx.core.custom_logging import iprint
from genx.gui.parametergrid import ValueCellRenderer

_set_func_prefix='set'

class SampleGrid(gridlib.Grid):
    def __init__(self, parent, *args, **kw):
        gridlib.Grid.__init__(self, parent, *args, **kw)
        self.parent=parent

        self.cb=None
        self.Bind(gridlib.EVT_GRID_CELL_LEFT_CLICK, self.onCellSelected)
        self.Bind(gridlib.EVT_GRID_EDITOR_CREATED, self.onEditorCreated)
        self.Bind(gridlib.EVT_GRID_EDITOR_SHOWN, self.onEditorShown)
        self.Bind(gridlib.EVT_GRID_EDITOR_HIDDEN, self.onEditorHidden)
        self._activated_ctrl=False

    def onCellSelected(self, evt):
        if evt.Col in [1]:
            self._activated_ctrl=True
            wx.CallAfter(self.EnableCellEditControl)
        if evt.Col in [3, 5, 7, 9]:
            if not self.parent.sample_table.GetAttr(evt.Row, evt.Col, None).IsReadOnly():
                self.parent.sample_table.SetValue(evt.Row, evt.Col,
                                                  not self.parent.sample_table.GetValue(evt.Row, evt.Col))
                self.ForceRefresh()
        else:
            evt.Skip()

    def onEditorCreated(self, evt):
        # Show tooltip on formula entry to give feedback on input
        inp=evt.Control
        inp.Bind(wx.EVT_TEXT, self.onFormula)
        if evt.Col in [3, 5, 7, 9] and self._activated_ctrl:
            self._activated_ctrl=False
        if evt.Col==2 and self.GetTable().GetValue(evt.Row, 1)=='Formula':
            self.info_text.Show()
            self.info_text.SetLabel('Enter Chemical Formula:')
            self.parent.Layout()
        evt.Skip()

    def onEditorHidden(self, evt):
        if self.info_text.IsShown():
            self.info_text.Hide()
            self.parent.Layout()
        evt.Skip()

    def onEditorShown(self, evt):
        if evt.Col==2 and self.GetTable().GetValue(evt.Row, 1)=='Formula':
            # Show tooltip on formula entry to give feedback on input
            self.info_text.Show()
            self.info_text.SetLabel('Enter Chemical Formula:')
            self.parent.Layout()
        evt.Skip()

    def onFormula(self, evt):
        if not self.info_text.IsShown():
            evt.Skip()
            return
        txt=evt.GetString()
        try:
            frm=Formula.from_str(txt)
        except Exception as e:
            self.info_text.SetLabel('Error in Formula:\n'+str(e))
        else:
            txt='Analyzed Formula:\n'+frm.describe()
            if frm in mdb:
                txt+='\n\nFound in DB:\n%g g/cm³'%mdb.dens_mass(frm)
            self.info_text.SetLabel(txt)

# new model is ready with a script as value.
(update_model_event, EVT_UPDATE_MODEL)=wx.lib.newevent.NewEvent()
TOP_LAYER=0
ML_LAYER=1
BOT_LAYER=2

class SampleTable(gridlib.GridTableBase):
    _columns=[
        ('Layer', gridlib.GRID_VALUE_STRING),
        ('Formula Params:\n'
         '---------------'
         '\nMixure Params:', gridlib.GRID_VALUE_CHOICE+':Formula,Mixure'),
        ('Chem. Formula\n'
         '-------------'
         '\nSLD-1 [10⁻⁶Å⁻²]', gridlib.GRID_VALUE_STRING),
        ('', gridlib.GRID_VALUE_BOOL, False),
        ('Density [g/cm³]\n'
         '---------------'
         '\nSLD-2 [10⁻⁶Å⁻²]', gridlib.GRID_VALUE_STRING),
        ('', gridlib.GRID_VALUE_BOOL, False),
        ('Moment [µB/FU]\n'
         '--------------'
         '\nFraction [% SLD-1]', gridlib.GRID_VALUE_STRING),
        ('', gridlib.GRID_VALUE_BOOL, True),
        ('d [Å]', gridlib.GRID_VALUE_STRING),
        ('', gridlib.GRID_VALUE_BOOL, False),
        ('σ [Å]', gridlib.GRID_VALUE_STRING),
        ]

    _last_layer_data=[]

    defaults={
        'Formula': ['Layer', 'Formula', 'SLD',
                    False, '2.0', False, '0.0',
                    True, '10.0', False, '5.0', ML_LAYER],
        'Mixure': ['MixLayer', 'Mixure', '6.0',
                   False, '2.0', False, '100',
                   True, '10.0', False, '5.0', ML_LAYER],
        }

    repetitions=1

    def __init__(self, parent, grid):
        gridlib.GridTableBase.__init__(self)
        self.parent=parent
        self.grid=grid

        self.ResetModel(first=True)

        self.grid.SetTable(self, True)

        dpi_scale_factor=wx.GetApp().dpi_scale_factor

        self.grid.SetRowLabelSize(30*dpi_scale_factor)
        self.grid.SetColLabelSize(50*dpi_scale_factor)
        for i, colinfo in enumerate(self._columns):
            # self.parent.SetColSize(i, 50)
            self.grid.AutoSizeColumn(i, True)
        wx.CallAfter(self.updateModel)

    def ResetModel(self, first=False):
        if not first:
            old_len=len(self.layers)

        self.ambient=[None, 'Formula', 'SLD',
                      False, '0.0', False, '0.0',
                      False, '0', False, '0']
        self.substrate=[None, 'Formula', Formula([['Si', 1.0]]),
                        False, '2.32998', False, '0.0',
                        False, '0', False, '5.0']
        self.layers=[['Surface_Layer', 'Formula', Formula([['Fe', 2.0], ['O', 2.0]]),
                      False, '5.25568', False, '0.0',
                      False, '20.0', False, '5.0', TOP_LAYER],
                     ['Layer_1', 'Formula', Formula([['Fe', 1.0]]),
                      False, '7.87422', False, '0.0',
                      False, '100.0', False, '5.0', ML_LAYER],
                     ['Interface_Layer', 'Formula', Formula([['Si', 1.0], ['O', 2.0]]),
                      False, '4.87479', False, '0.0',
                      False, '20.0', False, '5.0', BOT_LAYER]
                     ]

        if not first:
            diff=old_len-len(self.layers)
            iprint(diff, 'reset table')
            for i in range(abs(diff)):
                if diff<0:
                    msg=gridlib.GridTableMessage(self,
                                                 gridlib.GRIDTABLE_NOTIFY_ROWS_APPENDED, 1, 1)
                    self.GetView().ProcessTableMessage(msg)
                elif diff>0:
                    msg=gridlib.GridTableMessage(self,
                                                 gridlib.GRIDTABLE_NOTIFY_ROWS_DELETED, 1, 1)
                    self.GetView().ProcessTableMessage(msg)
            msg=gridlib.GridTableMessage(self,
                                         gridlib.GRIDTABLE_REQUEST_VIEW_GET_VALUES)
            self.GetView().ProcessTableMessage(msg)
            self.GetView().ForceRefresh()
            self.updateModel()

    def GetNumberRows(self):
        return len(self.layers)+3

    def GetNumberCols(self):
        return len(self._columns)

    def GetRowLabelValue(self, row):
        if row==self.repeatedInfo():
            return ''
        row=self.realRow(row)
        if row in [0, self.GetNumberRows()-2]:
            return '-'
        else:
            return '% 2i'%row

    def IsEmptyCell(self, row, col):
        try:
            return not self.GetValue(row, col)
        except IndexError:
            return True

    def GetValue(self, row, col):
        if row==self.repeatedInfo():
            if col==0:
                return 'Repeated layer structure'
            if col==8:
                return 'Repetitions:'
            if col==10:
                return str(self.repetitions)
            return None
        row=self.realRow(row)
        if col==0:
            if row==0:
                return 'Ambient'
            elif row==(self.GetNumberRows()-2):
                return 'Substrate'
            else:
                return self.layers[row-1][col].replace('_', ' ')
        if row==0:
            if col in [7, 8, 9, 10]:
                return None
            return self.ambient[col]
        elif row==self.GetNumberRows()-2:
            if col in [7, 8]:
                return None
            return self.substrate[col]

        return self.layers[row-1][col]

    def get_valid_name(self, name):
        # generate a valid identifier string from name
        identifyier=''
        for char in name.replace(' ', '_'):
            if (identifyier+char).isidentifier():
                identifyier+=char
        if identifyier in self.invalid_identifiers:
            identifyier='_'+identifyier

        existing=[li[0] for li in self.layers]
        if not identifyier in existing:
            return identifyier
        if identifyier.split('_')[-1].isdigit():
            identifyier=identifyier.rsplit('_', 1)[0]
        i=1
        while '%s_%i'%(identifyier, i) in existing:
            i+=1
        return '%s_%i'%(identifyier, i)

    def SetValue(self, row, col, value):
        # ignore unchanged values
        if value==self.GetValue(row, col):
            return

        if row==self.repeatedInfo():
            if col==10:
                self.repetitions=int(value)
            self.updateModel()
            return
        row=self.realRow(row)

        if row==0:
            to_edit=self.ambient
        elif row==(self.GetNumberRows()-2):
            to_edit=self.substrate
        else:
            to_edit=self.layers[row-1]
        if col==0:
            # name change
            old_name=to_edit[0]
            to_edit[0]='AboutToChangeValue'
            to_edit[0]=self.get_valid_name(value)
            self.delete_grid_items(old_name)
        elif col==2:
            # check formula
            if to_edit[1]=='Formula':
                if value=='SLD':
                    to_edit[2]=value
                else:
                    try:
                        formula=Formula.from_str(value)
                    except:
                        pass
                    else:
                        to_edit[2]=formula
                        # a new formula was set, if in DB, set its density
                        if formula in mdb:
                            to_edit[4]='%g'%mdb.dens_mass(formula)
            else:
                try:
                    val=float(eval('%s'%value))
                    if 0<=val<=100:
                        to_edit[2]=value
                except:
                    pass
                else:
                    to_edit[col]=value
        elif col==1:
            # change of layer type resets material data columns
            to_edit[1]=value
            for i in [2, 3, 4, 5, 6]:
                to_edit[i]=self.defaults[value][i]
        elif col in [3, 5, 7, 9]:
            # boolean columns are always correct
            to_edit[col]=value
        elif col in [4, 6, 8, 10]:
            # evaluate float values, can be written as formla
            try:
                float(eval('%s'%value))
            except:
                pass
            else:
                to_edit[col]=value
        self.updateModel()

    def updateModel(self, evt=None):
        model_code=self.getModelCode()
        evt=update_model_event()
        evt.script=model_code
        wx.PostEvent(self.parent, evt)

    def repeatedInfo(self):
        # get the first row of the repeated layer structure
        info_row=1
        for li in self.layers:
            if li[11]!=TOP_LAYER:
                break
            info_row+=1
        return info_row

    def realRow(self, row):
        info_row=self.repeatedInfo()
        if row<info_row:
            return row
        else:
            return row-1

    def GetAttr(self, row, col, kind):
        '''Called by the grid to find the attributes of the cell,
        bkg color, text colour, font and so on.
        '''
        attr=gridlib.GridCellAttr()
        if row==self.repeatedInfo():
            if col!=10:
                attr.SetReadOnly()
            if col==0:
                attr.SetSize(1, 8)
                attr.SetAlignment(wx.ALIGN_CENTER, wx.ALIGN_BOTTOM)
                return attr
            if col==8:
                attr.SetSize(1, 2)
                return attr
            if col==10:
                return attr
            return attr
        row=self.realRow(row)

        attr.SetAlignment(wx.ALIGN_LEFT, wx.ALIGN_CENTER)
        if row==0 and col==6:
            # ambiance has no thickness or roughness
            attr.SetSize(1, 5)
        if row==(self.GetRowsCount()-2) and col==6:
            # ambiance has no thickness
            attr.SetSize(1, 3)
        if row in [0, (self.GetRowsCount()-2)]:
            if row==0:
                if col==1:
                    attr.SetAlignment(wx.ALIGN_CENTER, wx.ALIGN_TOP)
                elif col in [3, 5, 7, 9]:
                    attr.SetAlignment(wx.ALIGN_RIGHT, wx.ALIGN_TOP)
                    # If layer is defined as fraction, only allow fitting of either
                    # density 2 or fraction.
                    if self.ambient[1]=='Mixure' and col==3 and self.ambient[5]:
                        attr.SetReadOnly()
                    elif self.ambient[1]=='Mixure' and col==5 and self.ambient[3]:
                        attr.SetReadOnly()
                else:
                    attr.SetAlignment(wx.ALIGN_LEFT, wx.ALIGN_TOP)
                attr.SetBackgroundColour('#dddddd')
                if col in [9, 10]:
                    attr.SetReadOnly()
            else:
                if col==1:
                    attr.SetAlignment(wx.ALIGN_CENTER, wx.ALIGN_BOTTOM)
                elif col in [3, 5, 7, 9]:
                    attr.SetAlignment(wx.ALIGN_RIGHT, wx.ALIGN_BOTTOM)
                    # If layer is defined as fraction, only allow fitting of either
                    # density 2 or fraction.
                    if self.substrate[1]=='Mixure' and col==3 and self.substrate[5]:
                        attr.SetReadOnly()
                    elif self.substrate[1]=='Mixure' and col==5 and self.substrate[3]:
                        attr.SetReadOnly()
                else:
                    attr.SetAlignment(wx.ALIGN_LEFT, wx.ALIGN_BOTTOM)
                attr.SetBackgroundColour('#aaaaff')
            if col in [0, 7, 8]:
                attr.SetReadOnly()
        else:
            if col==0:
                attr.SetAlignment(wx.ALIGN_RIGHT, wx.ALIGN_CENTER)
            if col==1:
                attr.SetAlignment(wx.ALIGN_CENTER, wx.ALIGN_CENTER)
            if col in [3, 5, 7, 9]:
                attr.SetAlignment(wx.ALIGN_RIGHT, wx.ALIGN_CENTER)
                # If layer is defined as fraction, only allow fitting of either
                # density 2 or fraction.
                if self.layers[row-1][1]=='Mixure' and col==3 and self.layers[row-1][5]:
                    attr.SetReadOnly()
                elif self.layers[row-1][1]=='Mixure' and col==5 and self.layers[row-1][3]:
                    attr.SetReadOnly()
            if self.layers[row-1][11]==TOP_LAYER:
                attr.SetBackgroundColour('#ccffcc')
            elif self.layers[row-1][11]==BOT_LAYER:
                attr.SetBackgroundColour('#ffaaff')
        if col in [4, 6, 8, 10]:
            attr.SetRenderer(ValueCellRenderer())
        return attr

    def GetColLabelValue(self, col):
        '''Called when the grid needs to display labels
        '''
        return self._columns[col][0]

    def GetTypeName(self, row, col):
        '''Called to determine the kind of editor/renderer to use by
        default, doesn't necessarily have to be the same type used
        natively by the editor/renderer if they know how to convert.
        '''
        return self._columns[col][1]

    def CanGetValueAs(self, row, col, type_name):
        '''Called to determine how the data can be fetched and stored by the
        editor and renderer.  This allows you to enforce some type-safety
        in the grid.
        '''
        col_type=self._columns[col][1].split(':')[0]
        if type_name==col_type:
            return True
        else:
            return False

    def CanSetValueAs(self, row, col, type_name):
        return self.CanGetValueAs(row, col, type_name)

    def SetParameters(self, pars, clear=True, permanent_change=True):
        '''
        SetParameters(self, pars) --> None

        Set the parameters in the table to pars.
        pars has to an instance of Parameters.
        '''
        pass

    def InsertRow(self, row):
        model_row=self.realRow(row)
        if model_row==(self.GetNumberRows()-2):
            layer_type=self.substrate[1]
            layer_stack=BOT_LAYER
            model_row-=1
        elif row==self.repeatedInfo():
            layer_type=self.layers[row-1][1]
            layer_stack=self.layers[row-1][11]
        elif model_row>0:
            layer_type=self.layers[model_row-1][1]
            layer_stack=self.layers[model_row-1][11]
        else:
            layer_type=self.ambient[1]
            layer_stack=TOP_LAYER
        newlayer=list(self.defaults[layer_type])
        newlayer[11]=layer_stack
        newlayer[0]=self.get_valid_name(newlayer[0])
        self.layers.insert(model_row, newlayer)

        msg=gridlib.GridTableMessage(self,
                                     gridlib.GRIDTABLE_NOTIFY_ROWS_APPENDED, 1, 1)
        self.GetView().ProcessTableMessage(msg)
        msg=gridlib.GridTableMessage(self,
                                     gridlib.GRIDTABLE_REQUEST_VIEW_GET_VALUES)
        self.GetView().ProcessTableMessage(msg)
        self.GetView().ForceRefresh()
        self.updateModel()
        return True

    def RebuildTable(self, data):
        diff=len(self.layers)-len(data)
        self.layers=list(data)

        for i in range(abs(diff)):
            if diff<0:
                # list of layers is longer than previous table data
                msg=gridlib.GridTableMessage(self,
                                             gridlib.GRIDTABLE_NOTIFY_ROWS_APPENDED,
                                             1, 1)
                self.GetView().ProcessTableMessage(msg)
            elif diff>0:
                # list of layers is shorter than previous table data
                msg=gridlib.GridTableMessage(self,
                                             gridlib.GRIDTABLE_NOTIFY_ROWS_DELETED,
                                             1, 1)
                self.GetView().ProcessTableMessage(msg)
        msg=gridlib.GridTableMessage(self,
                                     gridlib.GRIDTABLE_REQUEST_VIEW_GET_VALUES)
        self.GetView().ProcessTableMessage(msg)
        self.GetView().ForceRefresh()
        self.updateModel()

    def DeleteRow(self, row):
        if row==self.repeatedInfo():
            return False
        row=self.realRow(row)
        if row in [0, self.GetNumberRows()-1]:
            return False
        # make sure we don't delete the last ML layer
        if self.layers[row-1][11]==ML_LAYER and len([li for li in self.layers if li[11]==ML_LAYER])==1:
            return False
        del_layer=self.layers.pop(row-1)

        # remove grid fit parameters from model, if any
        grid_parameters=self.parent.plugin.GetModel().get_parameters()
        for fi in ['dens', 'magn', 'd', 'sigma']:
            func_name=del_layer[0]+'.'+_set_func_prefix+fi.capitalize()
            grid_parameters.set_fit_state_by_name(func_name, 0., 0, 0., 0.)
        self.parent.UpdateGrid(grid_parameters)

        msg=gridlib.GridTableMessage(self,
                                     gridlib.GRIDTABLE_NOTIFY_ROWS_DELETED, 1, 1)
        self.GetView().ProcessTableMessage(msg)
        msg=gridlib.GridTableMessage(self,
                                     gridlib.GRIDTABLE_REQUEST_VIEW_GET_VALUES)
        self.GetView().ProcessTableMessage(msg)
        self.GetView().ForceRefresh()
        self.updateModel()
        return True

    def MoveRow(self, row_from, row_to):
        ignore_rows=[0, self.repeatedInfo(), self.GetNumberRows()-1]
        if row_from in ignore_rows or row_to in ignore_rows or row_to<0:
            return False
        row_from=self.realRow(row_from)
        row_to=self.realRow(row_to)
        if self.layers[row_from-1][11]!=self.layers[row_to-1][11]:
            return False
        moved_row=self.layers.pop(row_from-1)
        self.layers.insert(row_to-1, moved_row)

        msg=gridlib.GridTableMessage(self,
                                     gridlib.GRIDTABLE_REQUEST_VIEW_GET_VALUES)
        self.GetView().ProcessTableMessage(msg)
        self.GetView().ForceRefresh()
        self.updateModel()
        return True

    def getLayerCode(self, layer):
        out_param={}
        output="model.Layer("
        if layer[1]=='Formula':
            formula=layer[2]
            if formula=='SLD':
                nSLD=float(eval(layer[4]))
                mSLD=float(eval(layer[6]))
                output+="f=%s, "%(10*nSLD-10j*mSLD)
                output+="b=%s, "%nSLD
                output+="dens=0.1, magn=%s, "%mSLD
                out_param['dens']=0.1
                out_param['magn']=mSLD
            else:
                dens=(eval(layer[4])*MASS_DENSITY_CONVERSION/formula.mFU())
                output+="f=%s, "%formula.f()
                output+="b=%s, "%formula.b()
                output+="dens=%s, "%dens
                output+="magn=%s, "%layer[6]
                out_param['dens']=dens
                out_param['magn']=float(eval(layer[6]))
        else:
            SLD1=float(eval(layer[2]))
            SLD2=float(eval(layer[4]))
            frac=float(eval(layer[6]))/100.
            output+="f=(%s*%s+(1-%s)*%s), "%(frac, SLD1, frac, SLD2)
            output+="b=(%s*%s+(1-%s)*%s), "%(frac, SLD1, frac, SLD2)
            output+="dens=0.1, magn=0.0, "
            out_param['dens']=0.1
            out_param['magn']=0.0
        output+="d=%s, "%layer[8]
        output+="sigma=%s, "%layer[10]
        output+="xs_ai=0.0, magn_ang=0.0)"
        out_param['d']=float(eval(layer[8]))
        out_param['sigma']=float(eval(layer[10]))
        return output, out_param

    invalid_identifiers=['sample', 'Sim', 'model',
                         'fw', 'fp', 'bc', 'bw',
                         'Amb', 'Sub', 'inst']

    def getModelCode(self):
        '''
        Generate the python code for the current sample structure.
        '''
        grid_parameters=self.parent.plugin.GetModel().get_parameters()

        script="# BEGIN Sample DO NOT CHANGE\n"
        li, oi=self.getLayerCode(self.ambient)
        script+="Amb = %s\n"%li
        for pi, fi in [(3, 'dens'), (5, 'magn')]:
            if pi==5 and self.ambient[1]=='Mixure':
                if not self.ambient[5]:
                    continue
                fi='dens'
            value=oi[fi]
            minval=value*0.5
            maxval=value*2.0
            func_name='Amb.'+_set_func_prefix+fi.capitalize()
            grid_parameters.set_fit_state_by_name(func_name, 0., 0, 0., 0.)
            grid_parameters.set_fit_state_by_name(func_name, value,
                                                  int(self.ambient[pi]), minval, maxval)

        for layer in self.layers:
            li, oi=self.getLayerCode(layer)
            script+="%s = %s\n"%(layer[0], li)
            for pi, fi in [(3, 'dens'), (5, 'magn'), (7, 'd'), (9, 'sigma')]:
                if pi==5 and layer[1]=='Mixure':
                    if not layer[5]:
                        continue
                    fi='dens'
                value=oi[fi]
                minval=value*0.5
                maxval=value*2.0
                func_name=layer[0]+'.'+_set_func_prefix+fi.capitalize()
                grid_parameters.set_fit_state_by_name(func_name, 0., 0, 0., 0.)
                grid_parameters.set_fit_state_by_name(func_name, value,
                                                      int(layer[pi]), minval, maxval)

        li, oi=self.getLayerCode(self.substrate)
        script+="\nSub = %s\n"%li
        for pi, fi in [(3, 'dens'), (5, 'magn'), (9, 'sigma')]:
            if pi==5 and self.substrate[1]=='Mixure':
                if not self.substrate[5]:
                    continue
                fi='dens'
            value=oi[fi]
            minval=value*0.5
            maxval=value*2.0
            func_name='Sub.'+_set_func_prefix+fi.capitalize()
            grid_parameters.set_fit_state_by_name(func_name, 0., 0, 0., 0.)
            grid_parameters.set_fit_state_by_name(func_name, value,
                                                  int(self.substrate[pi]), minval, maxval)

        self.parent.UpdateGrid(grid_parameters)

        top=[li[0] for li in self.layers if li[11]==TOP_LAYER]
        ml=[li[0] for li in self.layers if li[11]==ML_LAYER]
        bot=[li[0] for li in self.layers if li[11]==BOT_LAYER]
        script+="\nTop = model.Stack(Layers=[%s ], Repetitions = 1)\n"%str(
            ", ".join(reversed(top))
            )
        script+="\nML = model.Stack(Layers=[%s ], Repetitions = %i)\n"%(str(
            ", ".join(reversed(ml))), self.repetitions)
        script+="\nBot = model.Stack(Layers=[%s ], Repetitions = 1)\n"%str(
            ", ".join(reversed(bot))
            )
        script+="\nsample = model.Sample(Stacks = [Bot, ML, Top], Ambient = Amb, Substrate = Sub)\n" \
                "# END Sample\n\n" \
                "# BEGIN Parameters DO NOT CHANGE\n"
        # store data used for the last script for reuse on update
        self._last_layer_data=[list(self.ambient)]
        for li in self.layers:
            self._last_layer_data.append(list(li))
        self._last_layer_data.append(list(self.substrate))
        return script

    def delete_grid_items(self, name):
        # remove fit grid entries corresponding to a renamed layer
        grid_parameters=self.parent.plugin.GetModel().get_parameters()
        for fi in ['dens', 'magn', 'd', 'sigma']:
            func_name=name+'.'+_set_func_prefix+fi.capitalize()
            grid_parameters.set_fit_state_by_name(func_name, 0., 0, 0., 0.)
        self.parent.UpdateGrid(grid_parameters)

    def get_name_list(self):
        out=['Amb']
        for li in self.layers:
            out.append(li[0])
        out.append('Sub')
        return out

    def update_layer_parameters(self, layer, dens=None, magn=None,
                                d=None, sigma=None):
        # update the table during/after a fit, layer can be index or name
        if type(layer) is not int:
            layer=self.get_name_list().index(layer)
        if layer==0:
            data=self.ambient
        elif layer==(len(self.layers)+1):
            data=self.substrate
        else:
            data=self.layers[layer-1]
        ref_data=self._last_layer_data[layer]

        if data[1]=='Formula':
            formula=data[2]
            if formula=='SLD':
                if dens is not None and data[3]:
                    data[4]=str(eval(ref_data[4])*dens/0.1)
                if magn is not None and data[5]:
                    data[6]=str(eval(ref_data[6])*dens/0.1)
            else:
                if dens is not None and data[3]:
                    new_dens=float(dens)*formula.mFU()/MASS_DENSITY_CONVERSION
                    data[4]=str(new_dens)
                if magn is not None and data[5]:
                    data[6]=str(float(magn))
        elif dens is not None:
            # FIXIT: now yet working as calculation compares with current value
            SLD1=float(eval(ref_data[2]))
            SLD2=float(eval(ref_data[4]))
            frac=float(eval(ref_data[6]))/100.
            new_dens=(frac*SLD1+(1-frac)*SLD2)*dens/0.1
            if data[3]:
                # SLD-2 was fitted
                sld2_fraction=new_dens-frac*SLD1
                data[4]=str(sld2_fraction/(1.-frac))
            if data[5]:
                # percentage was fitted
                new_frac=(new_dens-SLD2)/(SLD1-SLD2)
                data[6]=str(new_frac*100.)
        if d is not None and data[7]:
            data[8]=str(float(d))
        if sigma is not None and data[9]:
            data[10]=str(float(sigma))

class SamplePanel(wx.Panel):
    last_sample_script=''

    inst_params=dict(probe='neutron pol',
                     I0=1.0,
                     res=0.0,
                     wavelength=1.54,
                     pol='uu',
                     Ibkg=0.,
                     samplelen=10.,
                     beamw=0.1,
                     footype='no corr',
                     name='inst',
                     coords='q')

    def __init__(self, parent, plugin, refindexlist=None):
        wx.Panel.__init__(self, parent)
        if refindexlist is None:
            refindexlist=[]
        self.refindexlist=refindexlist
        self.plugin=plugin
        self.variable_span=0.25
        self.inst_params=dict(SamplePanel.inst_params)

        # Colours indicating different states
        # Green wx.Colour(138, 226, 52), ORANGE wx.Colour(245, 121, 0)
        self.fit_colour=(245, 121, 0)
        # Tango Sky blue wx.Colour(52, 101, 164), wx.Colour(114, 159, 207)
        self.const_fit_colour=(114, 159, 207)

        boxver=wx.BoxSizer(wx.HORIZONTAL)
        boxhor=wx.BoxSizer(wx.VERTICAL)
        self.toolbar=wx.ToolBar(self, style=wx.TB_FLAT | wx.TB_HORIZONTAL)
        boxhor.Add((-1, 2))
        self.do_toolbar()
        boxhor.Add(self.toolbar, proportion=0, flag=wx.EXPAND, border=1)
        boxhor.Add((-1, 2))
        self.grid=SampleGrid(self, -1, style=wx.NO_BORDER)
        self.sample_table=SampleTable(self, self.grid)

        self.last_sample_script=''  # self.sample_table.getModelCode()
        self.Bind(EVT_UPDATE_MODEL, self.UpdateModel)
        boxhor.Add(self.grid, 1, wx.EXPAND)
        boxver.Add(boxhor, 1, wx.EXPAND)

        self.grid.info_text=wx.StaticText(self, -1, 'Bier')
        boxver.Add(self.grid.info_text, 0)
        self.grid.info_text.Hide()

        self.SetSizer(boxver)
        self.toolbar.Realize()
        self.update_callback=lambda event: ''
        self._last_grid_data=[]

    def do_toolbar(self):
        dpi_scale_factor=wx.GetApp().dpi_scale_factor
        tb_bmp_size=int(dpi_scale_factor*20)

        newid=wx.NewId()
        self.toolbar.AddTool(newid, 'Insert Layer',
                             bitmap=wx.Bitmap(images.insert_layer.GetImage().Scale(tb_bmp_size, tb_bmp_size)),
                             shortHelp='Insert a Layer')
        self.Bind(wx.EVT_TOOL, self.OnLayerAdd, id=newid)

        newid=wx.NewId()
        self.toolbar.AddTool(newid, 'Delete',
                             bitmap=wx.Bitmap(images.delete.GetImage().Scale(tb_bmp_size, tb_bmp_size)),
                             shortHelp='Delete item')
        self.Bind(wx.EVT_TOOL, self.OnLayerDelete, id=newid)

        newid=wx.NewId()
        self.toolbar.AddTool(newid, 'Move up',
                             bitmap=wx.Bitmap(images.move_up.GetImage().Scale(tb_bmp_size, tb_bmp_size)),
                             shortHelp='Move item up')
        self.Bind(wx.EVT_TOOL, self.MoveUp, id=newid)

        newid=wx.NewId()
        self.toolbar.AddTool(newid, 'Move down',
                             bitmap=wx.Bitmap(images.move_down.GetImage().Scale(tb_bmp_size, tb_bmp_size)),
                             shortHelp='Move item down')
        self.Bind(wx.EVT_TOOL, self.MoveDown, id=newid)
        self.toolbar.AddSeparator()

        newid=wx.NewId()
        button=wx.Button(self.toolbar, newid, label='Instrument Settings',
                         size=(132*dpi_scale_factor, 22*dpi_scale_factor))
        button.SetBitmap(wx.Bitmap(images.instrument.GetImage().Scale(tb_bmp_size, tb_bmp_size)), dir=wx.LEFT)
        self.toolbar.AddControl(button)
        self.Bind(wx.EVT_BUTTON, self.EditInstrument, id=newid)
        self.toolbar.AddSeparator()

        self.toolbar.AddStretchableSpace()

        newid=wx.NewId()
        button=wx.Button(self.toolbar, newid, label='to Advanced Modelling',
                         size=(150*dpi_scale_factor, 22*dpi_scale_factor))
        button.SetBitmap(wx.Bitmap(images.custom_parameters.GetImage().Scale(tb_bmp_size, tb_bmp_size)), dir=wx.LEFT)
        button.SetToolTip("Switch to Reflectivity plugin for advanced modeling options.\n"
                          "This converts the model and can't be undone.")
        self.toolbar.AddControl(button)
        self.Bind(wx.EVT_BUTTON, self.SwitchAdvancedReflectivity, id=newid)

    def ChangeRepetitions(self, evt):
        self.UpdateModel()

    def instrumentCode(self):
        pars=dict(self.inst_params)
        if pars['res']==0:
            pars['restype']='no conv'
        else:
            pars['restype']='full conv and varying res.'
        template="%(name)s = model.Instrument(probe='%(probe)s', wavelength=%(wavelength)g, " \
                 "coords='%(coords)s', I0=%(I0)g, res=%(res)g, " \
                 "restype='%(restype)s', respoints=11, " \
                 "resintrange=3, beamw=%(beamw)g, footype='%(footype)s', " \
                 "samplelen=%(samplelen)g, incangle=0.0, pol='%(pol)s', " \
                 "Ibkg=%(Ibkg)g, tthoff=0.0,)\n"
        if pars['probe'] in ['x-ray', 'neutron']:
            return ['inst'], template%pars
        output=template%pars
        if pars['probe']=='neutron pol':
            pars['pol']='dd'
            pars['name']='inst_down'
            output+=template%pars
            insts=['inst', 'inst_down']
        else:
            insts=['inst']
        return insts, output

    def UpdateModel(self, evt=None, first=False):
        coords=self.inst_params['coords']
        if evt in [None, 'inst']:
            sample_script=self.last_sample_script
        else:
            sample_script=evt.script
            self.last_sample_script=sample_script
        script='from numpy import *\n' \
               'import models.spec_nx as model\n' \
               'from models.utils import UserVars, fp, fw, bc, bw\n\n' \
               '# BEGIN Instrument DO NOT CHANGE\n' \
               'from models.utils import create_fp, create_fw\n'
        insts, inst_str=self.instrumentCode()
        script+=inst_str

        script+="inst_fp = create_fp(inst.wavelength); inst_fw = create_fw(inst.wavelength)\n" \
                "fp.set_wavelength(inst.wavelength); fw.set_wavelength(inst.wavelength)\n" \
                "# END Instrument\n\n"
        # add sample description code
        script+=sample_script
        script+="cp = UserVars()\n" \
                "# END Parameters\n\n" \
                "SLD = []\n" \
                "def Sim(data):\n" \
                "    I = []\n" \
                "    SLD[:] = []\n"
        datasets=self.model.data
        from genx import data
        if coords=='q':
            self.plugin.parent.plot_data.update_labels('q [Å$^{-1}$]')
            data.DataSet.simulation_params[0]=0.001
            data.DataSet.simulation_params[1]=0.601
        else:
            self.plugin.parent.plot_data.update_labels('2θ [°]')
            data.DataSet.simulation_params[0]=0.01
            data.DataSet.simulation_params[1]=6.01
        res_set=False
        for i, di in enumerate(datasets):
            di.run_command()
            script+="    # BEGIN Dataset %i DO NOT CHANGE\n" \
                    "    d = data[%i]\n"%(i,i)
            if hasattr(di, 'res') and di.res is not None:
                script+="    inst.setRes(data[%i].res)\n"%i
            elif res_set:
                script+="    inst.setRes(0.001)\n"
            script+="    I.append(sample.SimSpecular(d.x, %s))\n" \
                    "    if _sim: SLD.append(sample.SimSLD(None, None, inst))\n" \
                    "    # END Dataset %i\n"%(insts[i%len(insts)], i)
            if di.name.startswith('Data') and len(insts)>1:
                prefix=['Spin Up', 'Spin Down'][i%2]
                di.name=prefix+' %i'%(i//2+1)
                if i%2==0:
                    di.data_color=(0.7, 0.0, 0.0)
                    di.sim_color=(1.0, 0.0, 0.0)
                else:
                    di.data_color=(0.0, 0.0, 0.7)
                    di.sim_color=(0.0, 0.0, 1.0)
            elif (di.name.startswith('Spin') and len(insts)==1) or first:
                di.name='Data %i'%i
                di.data_color=(0.0, 0.7, 0.0)
                di.sim_color=(0.0, 1.0, 0.0)
        datasets.update_data()
        # TODO: this is a very ugly unstable solution, try to find alternative.
        self.plugin.parent.data_list.list_ctrl._UpdateImageList()

        script+="    return I"
        self.plugin.SetModelScript(script)

        if evt is not None and self.plugin.mb_autoupdate_sim.IsChecked():
            self.plugin.parent.eh_tb_simulate(None)

    def OnLayerAdd(self, evt):
        row=self.grid.GetGridCursorRow()
        self.sample_table.InsertRow(row)

    def OnLayerDelete(self, evt):
        row=self.grid.GetGridCursorRow()
        self.sample_table.DeleteRow(row)

    def SetUpdateCallback(self, func):
        ''' SetUpdateCallback(self, func) --> None

        Sets the update callback will be called when the sample is updated.
        The call is on the form func(event)
        '''
        self.update_callback=func

    def Update(self, update_script=True):
        if update_script:
            self.update_callback(None)

    def EditInstrument(self, evt):
        """Event handler that creates an dialog box to edit the instruments.

        :param evt:
        :return: Nothing
        """
        validators={
            'probe': ['x-ray', 'neutron', 'neutron pol'],
            'coords': ['q', '2θ'], 'I0': FloatObjectValidator(),
            'res': FloatObjectValidator(), 'wavelength': FloatObjectValidator(),
            'Ibkg': FloatObjectValidator(), 'samplelen': FloatObjectValidator(),
            'beamw': FloatObjectValidator(),
            'footype': ['no corr', 'gauss beam', 'square beam'],
            }
        inst_name='inst'
        vals={}
        editable={}
        grid_parameters=self.plugin.GetModel().get_parameters()

        pars=['probe', 'coords', 'wavelength', 'I0', 'Ibkg',
              'res', 'footype', 'samplelen', 'beamw']
        units={'probe': '', 'wavelength': 'Å', 'coords': '',
               'I0': 'arb.', 'res': '[coord]', 'beamw': 'mm',
               'footype': '', 'samplelen': 'mm',
               'Ibkg': 'arb.', '2θ': '°', 'q': 'Å$^-1$'}
        groups=[('Radiation', ('probe', 'wavelength', 'I0')),
                ('Data', ('coords', 'Ibkg', 'res')),
                ('Footprint Correction', ('footype', 'samplelen', 'beamw'))]
        for item in validators.keys():
            func_name=inst_name+'.'+_set_func_prefix+item.capitalize()
            editable[item]=grid_parameters.get_fit_state_by_name(func_name)
            val=self.inst_params[item]
            vals[item]=val

        dlg=ValidateFitDialog(self, pars, vals, validators,
                              title='Instrument Editor', groups=groups,
                              units=units,
                              editable_pars=editable)

        if dlg.ShowModal()==wx.ID_OK:
            vals=dlg.GetValues()
            states=dlg.GetStates()
            for key, value in vals.items():
                if type(self.inst_params[key]) is not str:
                    value=float(vals[key])
                    minval=value*0.5
                    maxval=value*2.0
                    func_name=inst_name+'.'+_set_func_prefix+key.capitalize()
                    grid_parameters.set_fit_state_by_name(func_name, value, states[key], minval, maxval)
                    self.inst_params[key]=value
                else:
                    self.inst_params[key]=value
            self.UpdateGrid(grid_parameters)
            self.UpdateModel(evt='inst')
        else:
            pass
        dlg.Destroy()

    def SwitchAdvancedReflectivity(self, evt):
        # Load Reflectivity plugin for more advanced options, this will unload current plugin
        plugin_control=self.plugin.parent.plugin_control

        try:
            plugin_control.plugin_handler.load_plugin('Reflectivity')
        except:
            outp=io.StringIO()
            traceback.print_exc(200, outp)
            tbtext=outp.getvalue()
            outp.close()
        else:
            plugin_control.RegisterPlugin('Reflectivity')

    def CheckGridUpdate(self, parameters=None):
        if parameters is None:
            new_grid=self.plugin.GetModel().get_parameters()
            new_data=[list(di) for di in new_grid.data]
        else:
            new_data=parameters
        if self._last_grid_data!=new_data:
            layers=self.sample_table.get_name_list()
            for (pi, val, _, _, _, _) in new_data:
                try:
                    name, param=pi.split('.', 1)
                except ValueError:
                    continue

                if name in layers:
                    if param==_set_func_prefix+'Dens':
                        self.sample_table.update_layer_parameters(name, dens=val)
                    if param==_set_func_prefix+'Magn':
                        self.sample_table.update_layer_parameters(name, magn=val)
                    if param==_set_func_prefix+'D':
                        self.sample_table.update_layer_parameters(name, d=val)
                    if param==_set_func_prefix+'Sigma':
                        self.sample_table.update_layer_parameters(name, sigma=val)
            msg=gridlib.GridTableMessage(self.sample_table,
                                         gridlib.GRIDTABLE_REQUEST_VIEW_GET_VALUES)
            self.sample_table.GetView().ProcessTableMessage(msg)
            self.sample_table.GetView().ForceRefresh()
            self._last_grid_data=new_data

    def UpdateGrid(self, grid_parameters):
        self._last_grid_data=[list(di) for di in grid_parameters.data]
        self.plugin.parent.paramter_grid.SetParameters(grid_parameters)

    def MoveUp(self, evt):
        row=self.grid.GetGridCursorRow()
        res=self.sample_table.MoveRow(row, row-1)
        if res:
            self.grid.MoveCursorUp(False)

    def MoveDown(self, evt):
        row=self.grid.GetGridCursorRow()
        res=self.sample_table.MoveRow(row, row+1)
        if res:
            self.grid.MoveCursorDown(False)

class WizarSelectionPage(WizardPageSimple):
    def __init__(self, parent, choices, intro_title='', intro_text='',
                 choice_label='', choices_help=None,
                 prev=None, next=None, bitmap=wx.NullBitmap):
        WizardPageSimple.__init__(self, parent, prev=prev, next=next, bitmap=bitmap)
        dpi_scale_factor=wx.GetApp().dpi_scale_factor

        vbox=wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(vbox)
        text=wx.StaticText(self, style=wx.ALIGN_LEFT, label=intro_title)
        text.SetFont(text.GetFont().Scale(1.5))
        vbox.Add(text, 0, wx.EXPAND)
        text=wx.StaticText(self, style=wx.ALIGN_LEFT, label=intro_text)
        text.Wrap(text.GetSize().width)
        vbox.Add(text, 0, wx.EXPAND)
        vbox.AddStretchSpacer()

        self.ctrl={}

        box=wx.StaticBox(self, label=choice_label)
        out_layout=wx.BoxSizer(wx.VERTICAL)
        box.SetSizer(out_layout)
        box_layout=wx.GridSizer(min(4, len(choices)//4+1), 2, 2)
        out_layout.Add(box_layout, 1, wx.EXPAND|wx.TOP|wx.LEFT, 12*dpi_scale_factor)
        for choice in choices:
            self.ctrl[choice]=wx.RadioButton(box, label=choice,
                                             size=wx.Size(-1, 16*dpi_scale_factor))
            box_layout.Add(self.ctrl[choice], 0, wx.FIXED_MINSIZE, 2)
        sz=box_layout.GetMinSize()
        sa=box.GetBordersForSizer()
        box.SetMinSize(wx.Size(sz.width+sa[0]+sa[1]+12*dpi_scale_factor, sz.height+sa[1]*2+12*dpi_scale_factor))

        #box.SetMinSize(wx.Size(500, 300))

        #wx.RadioBox(self, label=choice_label, choices=choices,
        #                      style=wx.RA_SPECIFY_ROWS, majorDimension=4)

        # dc=wx.ScreenDC()
        # dc.SetFont(self.ctrl.GetFont())
        # txt_height=dc.GetTextExtent("A").height
        # txt_lengths=[dc.GetTextExtent(ti).width for ti in choices]
        # total_txt_width=max(txt_lengths[:4])
        # for i in range(len(choices)//4):
        #     total_txt_width+=max(txt_lengths[i*4:(i+1)*4])
        # self.ctrl.SetMinSize(self.ctrl.GetBestSize())#wx.Size(total_txt_width, 50))

        vbox.Add(box, 0, 0)
        if choices_help:
            for i, ti in enumerate(choices_help):
                #self.ctrl.SetItemToolTip(i, ti)
                self.ctrl[choices[i]].SetToolTip(ti)
            text=wx.StaticText(self, style=wx.ALIGN_LEFT, label='Hover items for info')
            text.SetFont(text.GetFont().Scale(0.75))
            vbox.Add(text, 0, 0)

    def selection(self):
        #return self.ctrl.GetString(self.ctrl.GetSelection())
        for key, ctrl in self.ctrl.items():
            if ctrl.GetValue():
                return key

def find_code_segment(code, descriptor):
    '''find_code_segment(code, descriptor) --> string

    Finds a segment of code between BEGIN descriptor and END descriptor
    returns a LookupError if the segement can not be found
    '''
    found=0
    script_lines=code.splitlines(True)
    line_index=0
    for line in script_lines[line_index:]:
        line_index+=1
        if line.find('# BEGIN %s'%descriptor)!=-1:
            found+=1
            break

    text=''
    for line in script_lines[line_index:]:
        line_index+=1
        if line.find('# END %s'%descriptor)!=-1:
            found+=1
            break
        text+=line

    if found!=2:
        raise LookupError('Code segement: %s could not be found'%descriptor)

    return text

class Plugin(framework.Template):
    previous_xaxis=None
    sim_returns_sld=True

    def __init__(self, parent):
        if 'Reflectivity' in parent.plugin_control.plugin_handler.get_loaded_plugins():
            parent.plugin_control.UnLoadPlugin_by_Name('Reflectivity')
        framework.Template.__init__(self, parent)

        self.model_obj=self.GetModel()
        prev_script=self.model_obj.script
        sample_panel=self.NewInputFolder('Model')
        sample_sizer=wx.BoxSizer(wx.HORIZONTAL)
        sample_panel.SetSizer(sample_sizer)
        self.defs=['Instrument', 'Sample']
        self.sample_widget=SamplePanel(sample_panel, self)
        self.sample_widget.model=self.model_obj
        sample_sizer.Add(self.sample_widget, 1, wx.EXPAND | wx.GROW | wx.ALL)
        sample_panel.Layout()

        self.sample_widget.SetUpdateCallback(self.UpdateScript)

        # Create the SLD plot
        sld_plot_panel=self.NewPlotFolder('SLD')
        sld_sizer=wx.BoxSizer(wx.HORIZONTAL)
        sld_plot_panel.SetSizer(sld_sizer)
        self.sld_plot=SamplePlotPanel(sld_plot_panel, self)
        sld_sizer.Add(self.sld_plot, 1, wx.EXPAND | wx.GROW | wx.ALL)
        sld_plot_panel.Layout()

        # Create a menu for handling the plugin
        menu=self.NewMenu('Reflec')
        self.mb_export_sld=wx.MenuItem(menu, wx.NewId(),
                                       "Export SLD...",
                                       "Export the SLD to a ASCII file",
                                       wx.ITEM_NORMAL)
        menu.Append(self.mb_export_sld)
        self.mb_show_imag_sld=wx.MenuItem(menu, wx.NewId(),
                                          "Show Im SLD",
                                          "Toggles showing the imaginary part of the SLD",
                                          wx.ITEM_CHECK)
        menu.Append(self.mb_show_imag_sld)
        self.mb_show_imag_sld.Check(False)
        self.show_imag_sld=self.mb_show_imag_sld.IsChecked()
        self.mb_autoupdate_sim=wx.MenuItem(menu, wx.NewId(),
                                           "Simulate Autom.",
                                           "Simulates the model automatically after manual changes",
                                           wx.ITEM_CHECK)
        menu.Append(self.mb_autoupdate_sim)
        self.mb_autoupdate_sim.Check(True)
        self.mb_autoupdate_sld=wx.MenuItem(menu, wx.NewId(),
                                           "Autoupdate SLD",
                                           "Toggles autoupdating the SLD during fitting",
                                           wx.ITEM_CHECK)
        menu.Append(self.mb_autoupdate_sld)
        self.mb_autoupdate_sld.Check(True)

        self.mb_hide_advanced=wx.MenuItem(menu, wx.NewId(),
                                          "Hide Advanced",
                                          "Toggles hiding of advanced model tabs",
                                          wx.ITEM_CHECK)
        menu.Append(self.mb_hide_advanced)
        self.mb_hide_advanced.Check(True)

        if prev_script!='':
            if self.model_obj.filename!='':
                iprint("SimpleReflectivity plugin: Reading loaded model")
                self.ReadModel()
            else:
                try:
                    self.ReadModel()
                except:
                    iprint("SimpleReflectivity plugin: Creating new model")
                    self.OnNewModel(None)
        else:
            iprint("SimpleReflectivity plugin: Creating new model")
            self.OnNewModel(None)

        # self.mb_autoupdate_sld.SetCheckable(True)
        self.parent.Bind(wx.EVT_MENU, self.OnExportSLD, self.mb_export_sld)
        self.parent.Bind(wx.EVT_MENU, self.OnAutoUpdateSLD,
                         self.mb_autoupdate_sld)
        self.parent.Bind(wx.EVT_MENU, self.OnShowImagSLD,
                         self.mb_show_imag_sld)
        self.parent.Bind(wx.EVT_MENU, self.OnHideAdvanced,
                         self.mb_hide_advanced)

        self.DisableGrid()
        self.HideUIElements()
        self.sample_widget.UpdateModel()
        self.StatusMessage('Simple Reflectivity plugin loaded')

        self.parent.Bind(EVT_UPDATE_PARAMETERS, self.OnFitParametersUpdated)
        self.parent.Bind(EVT_PARAMETER_GRID_CHANGE, self.OnGridMayHaveErrors)

    def OnHideAdvanced(self, evt):
        if self.mb_hide_advanced.IsChecked():
            self.HideUIElements()
        else:
            self.ShowUIElements()

    def HideUIElements(self):
        self._hidden_pages=[]
        # hide all standard tabs
        nb=self.parent.input_notebook
        for i, page_i in reversed(list(enumerate(nb.Children))):
            title=nb.GetPageText(i)
            if title!='Model':
                self._hidden_pages.append([title, page_i, i])
                nb.RemovePage(i)
        self._hidden_pages.reverse()

    def ShowUIElements(self):
        nb=self.parent.input_notebook
        for title, page_i, i in self._hidden_pages:
            nb.InsertPage(i, page_i, title)
        self._hidden_pages=None

    def DisableGrid(self):
        nb=self.parent.input_notebook
        for i, page_i in enumerate(nb.Children):
            title=nb.GetPageText(i)
            if title=='Grid':
                page_i.Disable()

    def EnableGrid(self):
        nb=self.parent.input_notebook
        for i, page_i in enumerate(nb.Children):
            title=nb.GetPageText(i)
            if title=='Grid':
                page_i.Enable()

    def Remove(self):
        if self.mb_hide_advanced.IsChecked():
            self.ShowUIElements()
        self.EnableGrid()

        self.parent.Unbind(EVT_UPDATE_PARAMETERS, handler=self.OnFitParametersUpdated)
        self.parent.Unbind(EVT_PARAMETER_GRID_CHANGE, handler=self.OnGridMayHaveErrors)
        framework.Template.Remove(self)

    def UpdateScript(self, event):
        self.WriteModel()

    def OnAutoUpdateSLD(self, evt):
        # self.mb_autoupdate_sld.Check(not self.mb_autoupdate_sld.IsChecked())
        pass

    def OnShowImagSLD(self, evt):
        self.show_imag_sld=self.mb_show_imag_sld.IsChecked()
        self.sld_plot.Plot()

    def OnExportSLD(self, evt):
        dlg=wx.FileDialog(self.parent, message="Export SLD to ...",
                          defaultFile="",
                          wildcard="Dat File (*.dat)|*.dat",
                          style=wx.FD_SAVE | wx.FD_CHANGE_DIR
                          )
        if dlg.ShowModal()==wx.ID_OK:
            fname=dlg.GetPath()
            result=True
            if os.path.exists(fname):
                filepath, filename=os.path.split(fname)
                result=self.ShowQuestionDialog('The file %s already exists.'
                                               ' Do'
                                               ' you wish to overwrite it?'
                                               %filename)
            if result:
                try:
                    self.sld_plot.SavePlotData(fname)
                except IOError as e:
                    self.ShowErrorDialog(e.__str__())
                except Exception as e:
                    outp=io.StringIO()
                    traceback.print_exc(200, outp)
                    val=outp.getvalue()
                    outp.close()
                    self.ShowErrorDialog('Could not save the file.'
                                         ' Python Error:\n%s'%(val,))
        dlg.Destroy()

    def OnNewModel(self, event):
        ''' Create a new model
        '''
        dlg=self.parent.data_list.list_ctrl.data_loader_cont
        dls=list(sorted(dlg.plugin_handler.get_possible_plugins()))
        dls.remove('auto')

        wiz=Wizard(self.parent, title='Create New Model...',
                   bitmap=images.getinstrumentBitmap(), style=wx.DEFAULT_DIALOG_STYLE | wx.STAY_ON_TOP)

        # def show_dl_help(event):
        #     dlg=PluginHelpDialog(wiz, 'plugins.data_loaders')
        #     dlg.Show()
        p1=WizarSelectionPage(wiz, ['x-ray', 'neutron', 'neutron pol'],
                              'Select Probe\n',
                              'Please choose the experiment you want to model'
                              '/fit. This option can be changed later '
                              'from the Instrument Settings dialog.')
        p2=WizarSelectionPage(wiz, ['auto']+dls,
                              'Selecte Data Loader\n',
                              'How to import data into GenX. '
                              '\nIf your instrument is not listed '
                              'use the default/resolution loader. '
                              'This reads  3/4 columns from an ASCII file. '
                              'If the file does not '
                              'have x,y,dy(,res) column order, use Setting->'
                              'Import to select which columns to read.',
                              bitmap=getopenBitmap())
        p3=WizarSelectionPage(wiz, ['q', '2θ'],
                              'Select Data Coordinates\n',
                              'Set the x-coordinates used for simulation. '
                              '\nYou can define more experimental parameters like '
                              'wavelength, resolution, footprint correction or '
                              'background in the Instrument Settings dialog.',
                              choices_help=['Reciprocal lattice vector out-of-plane component qz in Å⁻¹',
                                            'Detector angle in degrees'],
                              bitmap=getplottingBitmap())
        WizardPageSimple.Chain(p1, p2)
        WizardPageSimple.Chain(p2, p3)

        dpi_scale_factor=wx.GetApp().dpi_scale_factor
        wiz.SetPageSize(wx.Size(int(400*dpi_scale_factor), int(240*dpi_scale_factor)))

        if wiz.RunWizard(p1):
            self.sample_widget.inst_params=dict(SamplePanel.inst_params)
            params=self.sample_widget.inst_params
            params['probe']=p1.selection()
            params['coords']=p3.selection()
            dlg.LoadPlugin(p2.selection())
            # dlg.plugin_handler.load_plugin(p2.selection())
            self.sample_widget.sample_table.ResetModel()
            self.sample_widget.UpdateModel(first=True)
        wiz.Destroy()

    def OnDataChanged(self, event):
        ''' Take into account changes in data..
        '''
        self.sample_widget.UpdateModel()

    def OnOpenModel(self, event):
        '''
        Loads the sample into the plugin...
        '''
        pass  # self.ReadModel()

    def OnSimulate(self, event):
        '''
        Updates stuff after simulation
        '''
        # Calculate and update the sld plot
        self.sld_plot.Plot()

    def OnFittingUpdate(self, event):
        '''
        Updates stuff during fitting
        '''
        # Calculate and update the sld plot
        if self.mb_autoupdate_sld.IsChecked():
            wx.CallAfter(self.sld_plot.Plot)

    def OnGridChange(self, event):
        self.sample_widget.CheckGridUpdate()
        self.sample_widget.Update(update_script=False)

    def OnGridMayHaveErrors(self, event):
        errors=[pi.error for pi in self.model_obj.parameters if pi.fit]
        if len(errors)>0 and not '-' in errors:
            #  summarize error information on fit parameters
            error_data=[]
            ST=self.sample_widget.sample_table
            layers=ST.get_name_list()
            for pi in self.model_obj.parameters:
                try:
                    name, param=pi.name.split('.', 1)
                except ValueError:
                    continue
                show_data=['', '', '', '']
                if name in layers:
                    lidx=ST.get_name_list().index(name)
                    if lidx==0:
                        st_data=ST.ambient
                        name='Ambient'
                    elif lidx==(len(ST.layers)+1):
                        name='Substrate'
                        st_data=ST.substrate
                    else:
                        st_data=ST.layers[lidx-1]
                    if param==_set_func_prefix+'Dens':

                        if st_data[1]=='Formula':
                            formula=st_data[2]
                            if formula=='SLD':
                                # result is SLD
                                show_data=[name, 'SLD [10^6 Å^-1]', '%.6g'%pi.value, pi.error]
                            else:
                                emin,emax=map(float, pi.error.strip('(').strip(')').strip().split(','))
                                scale=formula.mFU()/MASS_DENSITY_CONVERSION
                                show_data=[name, 'density [g/cm³]', '%.6g'%(pi.value*scale),
                                           '(%.4g, %.4g)'%(emin*scale, emax*scale)]
                        else:
                            show_data=[name, 'density', '%.6g'%pi.value, pi.error]
                    elif param==_set_func_prefix+'Magn':
                        show_data=[name, 'magnetization', '%.6g'%pi.value, pi.error]
                    elif param==_set_func_prefix+'D':
                        show_data=[name, 'thickness [Å]', '%.6g'%pi.value, pi.error]
                    elif param==_set_func_prefix+'Sigma':
                        show_data=[name, 'roughness [Å]', '%.6g'%pi.value, pi.error]
                    else:
                        show_data=[name, param, '%.6g'%pi.value, pi.error]
                else:
                    show_data=[pi.name, '', '%.6g'%pi.value, pi.error]
                error_data.append(show_data)
            dia=wx.Dialog(self.parent, style=wx.DEFAULT_DIALOG_STYLE | wx.RESIZE_BORDER | wx.MAXIMIZE_BOX)
            dia.SetTitle('Parameter Error Estimation')
            box=wx.BoxSizer(wx.VERTICAL)
            dia.SetSizer(box)
            grid=wx.grid.Grid(dia)
            grid.CreateGrid(len(error_data), 4)
            grid.DisableCellEditControl()
            grid.SetColLabelValue(0, 'Item')
            grid.SetColLabelValue(1, 'Parameter')
            grid.SetColLabelValue(2, 'Value')
            grid.SetColLabelValue(3, 'Error min/max')
            for i, row in enumerate(error_data):
                for j, cell_value in enumerate(row):
                    grid.SetCellValue(i, j, cell_value)
            for i in range(4):
                grid.AutoSizeColumn(i)
            box.Add(grid, proportion=1, flag=wx.EXPAND)

            w=min(grid.GetBestSize().width+80, self.parent.GetSize().width//1.5)
            h=min(grid.GetBestSize().height+80, self.parent.GetSize().height//1.3)
            dia.SetSize(wx.Size(w, h))

            dia.ShowModal()
        event.Skip()

    def OnFitParametersUpdated(self, event):
        grid_parameters=self.GetModel().get_parameters()
        keys=grid_parameters.get_fit_pars()[1]
        values=event.values
        parameters=[(key, value, None, None, None, None) for key, value in zip(keys, values)]
        self.sample_widget.CheckGridUpdate(parameters=parameters)
        self.sample_widget.Update(update_script=False)
        if event.permanent_change:
            for pi, val in zip(keys, values):
                try:
                    name, param=pi.split('.', 1)
                except ValueError:
                    continue
                if name=='inst':
                    for key in self.sample_widget.inst_params.keys():
                        if param==_set_func_prefix+key.capitalize():
                            value=float(val)
                            self.sample_widget.inst_params[key]=value
                            minval=value*0.5
                            maxval=value*2.0
                            grid_parameters.set_fit_state_by_name(pi, value, 0, 0, 0)
                            grid_parameters.set_fit_state_by_name(pi, value, 1, minval, maxval)
            self.sample_widget.sample_table.updateModel()

    def WriteModel(self):
        return

    def ReadModel(self):
        '''ReadModel(self)  --> None

        Reads in the current model and locates layers and stacks
        and sample defined inside BEGIN Sample section.
        '''
        self.StatusMessage('Compiling the script...')
        try:
            self.CompileScript()
        except GenxError as e:
            self.ShowErrorDialog(str(e))
            self.StatusMessage('Error when compiling the script')
            return
        except Exception as e:
            outp=io.StringIO()
            traceback.print_exc(200, outp)
            val=outp.getvalue()
            outp.close()
            self.ShowErrorDialog(val)
            self.Statusmessage('Fatal Error - compling, SimpleReflectivity')
            return
        self.StatusMessage('Script compiled!')

        self.StatusMessage('Trying to interpret the script...')

        txt=self.GetModel().script
        grid_parameters=self.GetModel().get_parameters()
        # extract instrument parameters
        insttxt=find_code_segment(txt, 'Instrument')
        for li in insttxt.splitlines():
            if not li.split('=', 1)[0].strip()=='inst':
                continue
            instoptions=li.split('model.Instrument(')[1].strip()[:-1]
            items=instoptions.strip(',').split(',')
            items=dict([tuple(map(str.strip, i.split('=', 1))) for i in items])
        for key, value in self.sample_widget.inst_params.items():
            if key in items:
                newval=type(value)(eval(items[key]))
                self.sample_widget.inst_params[key]=newval

        # extract layer parameters
        sampletxt=find_code_segment(txt, 'Sample')
        layers={}
        layer_order=[]
        stacks={}
        repetitions=1
        for li in sampletxt.splitlines():
            if 'model.Layer' in li:
                name, ltxt=map(str.strip, li.split('=', 1))
                layers[name]=analyze_layer_txt(name, ltxt)
                layer_order.append(name)
            if 'model.Stack' in li:
                name, stxt=map(str.strip, li.split('=', 1))
                stacks[name]=analyze_stack_txt(stxt)
                if name=='ML':
                    repetitions=int(stxt.split('Repetitions', 1)[1].split('=', 1)[1].split(',', 1)[0].rstrip(')'))
        if 'Top' in stacks:
            Top=stacks['Top']
        else:
            Top=[]
        if 'Bot' in stacks:
            Bottom=stacks['Bot']
        else:
            Bottom=[]
        for ni, di in layers.items():
            if ni in Top:
                di.append(TOP_LAYER)
            elif ni in Bottom:
                di.append(BOT_LAYER)
            else:
                di.append(ML_LAYER)

        # get parameters to be fitted
        for li in layers.values():
            prefix=li[0]+'.set'
            fit_items=[(3, 'dens'), (5, 'magn'), (7, 'd'), (9, 'sigma')]
            for index, si in fit_items:
                if grid_parameters.get_fit_state_by_name(prefix+si.capitalize()):
                    li[index]=True

        table=self.sample_widget.sample_table
        table.ambient=layers['Amb']
        table.ambient[0]=None
        table.substrate=layers['Sub']
        table.substrate[0]=None
        table.repetitions=repetitions
        layers=[layers[key] for key in layer_order if not key in ['Amb', 'Sub']]

        table.RebuildTable(layers)

        self.StatusMessage('New sample loaded to plugin!')

def analyze_layer_txt(name, txt):
    # ['Iron', 'Formula', Formula([['Fe', 1.0]]), False, '7.87422', False, '3.0', True, '100.0', False, '5.0']
    output=[name]
    layeroptions=txt.split('model.Layer(')[1].strip()[:-1]
    items=layeroptions.strip(',').split(',')
    items=dict([tuple(map(str.strip, i.split('=', 1))) for i in items])

    if 'bc.' in items['b']:
        output.append('Formula')
        output.append(Formula.from_bstr(items['b']))
        # density
        output.append(False)
        dens=float(items['dens'])*output[2].mFU()/MASS_DENSITY_CONVERSION
        output.append(str(dens))
        # magnetization
        output.append(False)
        output.append(str(eval(items['magn'])))
    elif '+' in items['b'] and '*' in items['b']:
        output.append('Mixure')
        part1, part2=items['b'].strip('()').split('+', 1)
        frac=float(part1.split('*')[0])*100
        SLD1=float(part1.split('*')[1])
        SLD2=float(part2.split('*')[1])
        output.append(str(SLD1))
        output.append(False)
        output.append(str(SLD2))
        output.append(False)
        output.append(str(frac))
    else:
        output.append('Formula')
        output.append('SLD')
        # density
        output.append(False)
        output.append(str(eval(items['b'])))
        # magnetization
        output.append(False)
        output.append(str(eval(items['magn'])))

    # thickness
    output.append(False)
    output.append(str(eval(items['d'])))
    # roughness
    output.append(False)
    output.append(str(eval(items['sigma'])))
    return output

def analyze_stack_txt(txt):
    txt=txt.split('Layers', 1)[1].split(']', 1)[0].strip('=[ ')
    layers=map(str.strip, txt.split(','))
    return list(layers)
